# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import warnings
from typing import Any, Iterable, Optional

import dace
from dace import (
    properties as dace_properties,
    subsets as dace_sbs,
    transformation as dace_transformation,
)
from dace.sdfg import graph as dace_graph, nodes as dace_nodes
from dace.transformation.passes import analysis as dace_analysis

from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations
from gt4py.next.program_processors.runners.dace.transformations import (
    splitting_tools as gtx_dace_split,
)


def gt_split_access_nodes(
    sdfg: dace.SDFG,
    validate: bool = True,
    validate_all: bool = False,
    single_use_data: Optional[dict[dace.SDFG, set[str]]] = None,
) -> Optional[int]:
    """Applies the `SplitAccessNode` transformation to the SDFG.

    This function should be the preferred way to run the `SplitAccessNode`
    transformation. Since it will ensure that the single data is only computed
    once. Furthermore, it guarantees that the transformations are applied in
    a deterministic order.

    The transformation returns the number of AccessNodes that have been split.

    Args:
        sdfg: The SDFG to process.
        validate: Perform validation after the pass has run.
        validate_all: Perform extensive validation.
        single_use_data: Which data descriptors are used only once.
            If not passed the function will run `FindSingleUseData`, if passed the
            function will update its content and add the newly generated data.
    """

    # To ensures that the `{src,dst}_subset` are properly set, run initialization.
    #  See [issue 1703](https://github.com/spcl/dace/issues/1703)
    for state in sdfg.states():
        for edge in state.edges():
            edge.data.try_initialize(sdfg, state, edge)

    if single_use_data is None:
        find_single_use_data = dace_analysis.FindSingleUseData()
        single_use_data = find_single_use_data.apply_pass(sdfg, None)

    apply_count = 0
    for nsdfg in sdfg.all_sdfgs_recursive():
        apply_count += _apply_split_access_node_non_recursive(
            sdfg=nsdfg,
            validate=validate,
            validate_all=validate_all,
            single_use_data=single_use_data[nsdfg],
        )

    return apply_count


def _apply_split_access_node_non_recursive(
    sdfg: dace.SDFG,
    validate: bool,
    validate_all: bool,
    single_use_data: set[str],
) -> int:
    apply_count = 0
    if len(single_use_data) == 0:
        return apply_count

    # The splitter transformation. Note that we set `assume_single_use_data` to `True`
    #  because we do this test outside.
    access_node_splitter = gtx_transformations.SplitAccessNode(
        assume_single_use_data=True,
    )

    # The transformation only applies to transient single use data, thus the order
    #  in which the states are provided are irrelevant. Furthermore, all fragments
    #  that are generated does not need to be examined again, thus a one pass is
    #  enough. This is because the `SplitAccessNode` transformation does not generates
    #  new edges upon the split, it just reroutes them to the fragments.
    for state in sdfg.states():
        state_cfg_id = state.parent_graph.cfg_id
        state_id = state.block_id
        scope_dict = state.scope_dict()

        # We can only split single use data that is also transient. Because GT4Py uses
        #  an SSA style we know that there is only one AccessNode that refers to that
        #  data. Thus, all AccessNodes that are stored refer to different data
        access_nodes_to_process = sorted(
            (
                dnode
                for dnode in state.data_nodes()
                if dnode.data in single_use_data and scope_dict[dnode] is None
            ),
            key=lambda dnode: dnode.data,
        )
        assert len(access_nodes_to_process) == len(
            set(map(lambda ac: ac.data, access_nodes_to_process))
        )

        if len(access_nodes_to_process) == 0:
            # Nothing to process in this state, continue.
            continue

        # Now try to split all candidates that we have found.
        for access_node_to_process in access_nodes_to_process:
            access_node_splitter.setup_match(
                sdfg=sdfg,
                cfg_id=state_cfg_id,
                state_id=state_id,
                subgraph={gtx_transformations.SplitAccessNode.access_node: access_node_to_process},
                expr_index=0,
                override=True,
            )
            if access_node_splitter.can_be_applied(
                graph=state, expr_index=0, sdfg=sdfg, permissive=False
            ):
                splitted_access_nodes = access_node_splitter.apply(graph=state, sdfg=sdfg)
                if validate_all:
                    # Not super correct as we not check at the top of the hierarchy.
                    sdfg.validate()

                # We have to update `single_use_data`. By definition all data that we
                #  generate through splitting is also single use data.
                single_use_data.update(sac.data for sac in splitted_access_nodes.values())
                apply_count += 1

    if validate:
        sdfg.validate()

    return apply_count


@dace_properties.make_properties
class SplitAccessNode(dace_transformation.SingleStateTransformation):
    """The transformation will split an AccessNode into multiple ones.

    If there is no interesection between a write and different reads,
    i.e. if every read to the AccessNode can be satisfied by a single
    write to the AccessNode and the AccessNode is only used at one
    location, then the node is split.
    This means that the reads will be satisfied directly and the node
    does not have to materialize.

    Before this transformation is run the `SplitConsumerMemlet` should
    be run.

    Args:
        single_use_data: The list of data that is used only once.
        assume_single_use_data: Assume that `access_node` is single use data.
            Note this flag should _only_ be used if the transformation is used
            through the `apply_to()` interface and the caller has ensured that
            `self.access_node` really is single use data.

    Todo:
        - Made it possible to merge splits, i.e. such that a fragment can
            be described by two producers.
        - Create a version that is able to split over multiple states. This is
            mostly useful to enable more state fusion.

    Note:
        The actual split operation is performed using `splitting_tools.split_node()`.
        Furthermore, as a special extension, to support certain workflows the
        `apply()` function returns the return value of that function.
    """

    access_node = dace_transformation.PatternNode(dace_nodes.AccessNode)

    assume_single_use_data = dace_properties.Property(
        dtype=bool,
        default=False,
        desc="Always assume that `self.access_node` is single use data. Only useful if used through `SplitAccessNode.apply_to()`.",
    )

    # Name of all data that is used at only one place. Is computed by the
    #  `FindSingleUseData` pass and be passed at construction time. Needed until
    #  [issue#1911](https://github.com/spcl/dace/issues/1911) has been solved.
    _single_use_data: Optional[dict[dace.SDFG, set[str]]]

    def __init__(
        self,
        *args: Any,
        single_use_data: Optional[dict[dace.SDFG, set[str]]] = None,
        assume_single_use_data: Optional[bool] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._single_use_data = single_use_data
        if assume_single_use_data is not None:
            self.assume_single_use_data = assume_single_use_data

    @classmethod
    def expressions(cls) -> Any:
        return [dace.sdfg.utils.node_path_graph(cls.access_node)]

    def can_be_applied(
        self,
        graph: dace.SDFGState,
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        access_node: dace_nodes.AccessNode = self.access_node
        desc = access_node.desc(sdfg)

        # To get rid of the intermediate it must be a single use transient.
        #  We postpone the single use check.
        if not desc.transient:
            return False
        if gtx_transformations.utils.is_view(desc, sdfg):
            return False

        # There must be multiple producers, otherwise this transformation
        #  does not make sense.
        number_of_producers = graph.in_degree(access_node)
        if number_of_producers <= 1:
            return False

        # Since this transformation can not handle splitting over multiple state
        #  it must be consumed directly, although single use data implies this
        #  we do it here explicitly to avoid a scan.
        #  It is also important that we explicitly allow one consumer. This case
        #  might imply that we have dead data.
        number_of_consumers = graph.out_degree(access_node)
        if number_of_consumers == 0:
            return False

        # Now check if a decomposition exist.
        edge_reassignments = self._find_edge_reassignment(graph)
        if edge_reassignments is None:
            return False
        if not self._check_split_constraints(
            state=graph,
            sdfg=sdfg,
            edge_reassignments=edge_reassignments,
        ):
            return False

        if self.assume_single_use_data:
            single_use_data = {sdfg: {access_node.data}}
        if self._single_use_data is None:
            find_single_use_data = dace_analysis.FindSingleUseData()
            single_use_data = find_single_use_data.apply_pass(sdfg, None)
        else:
            single_use_data = self._single_use_data
        if access_node.data not in single_use_data[sdfg]:
            return False

        return True

    def apply(
        self,
        graph: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> dict[dace_sbs.Subset, dace_nodes.AccessNode]:
        access_node: dace_nodes.AccessNode = self.access_node

        edge_reassignments = self._find_edge_reassignment(graph)
        assert edge_reassignments is not None

        # TODO(phimuell): Make it more general that it can take the full advantage
        #   of the splitter functionality.
        split_description = [e.data.dst_subset for e in edge_reassignments.keys()]

        fragment_access_nodes = gtx_dace_split.split_node(
            state=graph,
            sdfg=sdfg,
            node_to_split=access_node,
            split_description=split_description,
            allow_to_bypass_nodes=True,
        )

        # Split node will remove the AccessNode but does not remove the data.
        sdfg.remove_data(access_node.data, validate=False)

        # We have to clean up the isolated fragments. This is because we specified
        #  `allow_to_bypass_nodes` in the call above.
        for split_sbs in list(fragment_access_nodes.keys()):
            ac = fragment_access_nodes[split_sbs]
            if graph.degree(ac) == 0:
                graph.remove_node(ac)
                sdfg.remove_data(ac.data, validate=False)
                fragment_access_nodes.pop(split_sbs)

        # NOTE: In some situation it happens that when a producer writes
        #   something inside `access_node` and the data is never read. This is
        #   not an error, but can be a side effect of MapFusion or similar
        #   transformations. This will lead to dead data flow, that we will
        #   not remove. Instead DDE should be run.

        # Special extension to support certain workflows.
        return fragment_access_nodes

    def _find_edge_reassignment(
        self,
        state: dace.SDFGState,
    ) -> dict[dace_graph.MultiConnectorEdge, set[dace_graph.MultiConnectorEdge]] | None:
        """Determine how the edges should be distributed to the fragments.

        The current implementation defines the fragments, i.e. the pieces into
        which `self.access_node` should be split into, through the incoming edges.
        This means that every incoming edge defines one fragment.
        Therefor, the function returns a `dict` that maps each incoming edge to the
        set of out going edges that are associated to it, i.e. depend on the producer
        edge or `None` if such a distribution does not exist.

        The function does not perform any checks if the split lead to a valid
        SDFG, for that reason the result should be checked by
        `_check_split_constraints()`.

        Todo:
            Extend the function such that a fragment is not limited to a single
            incoming edge, but can also be defined through multiple edges.
        """
        access_node: dace_nodes.AccessNode = self.access_node

        # NOTE: Build the assignments based on the producers. Basing the split on
        #  the producers has the advantages that it naturally takes "double use"
        #  into account, i.e. one producer computes something and two different
        #  reads it. However, it is also quite unnatural, because the consumer
        #  should define the split (i.e. more than one producer are needed to
        #  generate the data for a consumer). This is hard to handle, but should
        #  be implemented at some point.
        edge_reassignments: dict[
            dace_graph.MultiConnectorEdge, set[dace_graph.MultiConnectorEdge]
        ] = {}
        for iedge in state.in_edges(access_node):
            if iedge.data.dst_subset is None:
                return None  # TODO(phimuell): Lift this.
            if iedge.data.wcr is not None:
                return None
            edge_reassignments[iedge] = set()

        # Now match the outgoing edges to their incoming producers.
        for oedge in state.out_edges(access_node):
            if oedge.data.wcr is not None:
                return None
            possible_producer = self._find_producer(oedge, edge_reassignments.keys())
            if possible_producer is None:
                return None
            edge_reassignments[possible_producer].add(oedge)

        unused_producers = [
            producer
            for producer, assigned_consumers in edge_reassignments.items()
            if len(assigned_consumers) == 0
        ]
        if unused_producers:
            # This situation is generated by MapFusion, if the intermediate
            #  AccessNode has to be kept alive.
            warnings.warn(
                "'SplitAccessNode': found producers "
                + ", ".join((str(p) for p in unused_producers))
                + " that generates data but that is never read.",
                stacklevel=0,
            )

        return edge_reassignments

    def _find_producer(
        self,
        consumer_edge: dace_graph.MultiConnectorEdge,
        producer_edges: Iterable[dace_graph.MultiConnectorEdge],
    ) -> dace_graph.MultiConnectorEdge | None:
        """Find the producer edge that generates what the consumer reads.

        The function checks which producer covers what the consumer reads.
        If there is not producer that does this, this function returns `None`.
        This function does not perform any additional tests.

        Args:
            consumer_edge: The edge that reads from `self.access_node`.
            producer_edges: List of all edges that writes to `self.access_node`.
        """
        consumer_subset = consumer_edge.data.src_subset

        # The consumer subset does not exist, so we can not do the decomposition.
        # TODO(phimuell): Fix this.
        if consumer_subset is None:
            return None

        # This check only checks if that the producer really generates the data that
        #  is consumed later. However, we also have to ensure that nothing is computed
        #  what is not consumed later. Thus.
        possible_producers = [
            producer_edge
            for producer_edge in producer_edges
            if producer_edge.data.dst_subset.covers(consumer_subset)
        ]

        # We only allow the case that one producer covers the consumer. If we found
        #  multiple candidates then we have an invalid SDFG, because multiple
        #  producer writes to the same memory location.
        if len(possible_producers) == 0:
            return None
        elif len(possible_producers) != 1:
            # This might indicate an error (the same memory location is written by
            #  multiple producer. However, there are some cases where it is not an
            #  error. For example a Map, with two Tasklets, writes to the node,
            #  one Tasklet writes `T[__i, 0]` the other `T[__i, 10]`, where `__i`
            #  is the iteration index. Then Memlet propagation will set the subset
            #  to something like `T[:, 0:10]`. So it is not an error in that case.
            warnings.warn(
                f"Found transient '{self.access_node.data}' that has multiple overlapping"
                " incoming edges. Might indicate an error.",
                stacklevel=0,
            )
            return None

        return possible_producers[0]

    def _check_split_constraints(
        self,
        state: dace.SDFGState,
        sdfg: dace.SDFG,
        edge_reassignments: dict[dace_graph.MultiConnectorEdge, set[dace_graph.MultiConnectorEdge]],
    ) -> bool:
        """Checks if the decomposition results in a valid SDFG.

        This function is used to validate the decomposition computed by
        `self._find_edge_reassignment()`.
        """

        for producer_edge, consumer_edges in edge_reassignments.items():
            data_source = producer_edge.src

            if len(consumer_edges) == 0:
                continue
            elif isinstance(data_source, dace_nodes.AccessNode):
                # TODO(phimuell): Should we also ensure that the domains are tight?
                if gtx_transformations.utils.is_view(data_source, sdfg):
                    return False

                # If the source is a global data, then we do not impose any other
                #  constraints.
                if not data_source.desc(sdfg).transient:
                    continue

                # If the source is a transient then we distinguish between two cases.
                #  In the first case there is only one consumer, in that case we
                #  require that everything is read. In the second case, more than
                #  one consumer, we do not impose any constraints.
                #  We do this to ensure the tightness of the temporaries, i.e. what
                #  is computed is also read, which is core assumption of the
                #  `CopyChainRemover`.
                # TODO(phimuell): Lift this limitation.
                if len(consumer_edges) == 1:
                    if not next(iter(consumer_edges)).data.src_subset.covers(
                        producer_edge.data.dst_subset
                    ):
                        return False

            elif isinstance(data_source, dace_nodes.MapExit):
                # The source is a Map, in this case we just generate a new transient
                #  output and then perform some reconnection. However, we require that
                #  all consumer read exactly what is is written by the map. This
                #  is to ensure some tightness of the domains.
                if not all(
                    consumer_edge.data.src_subset.covers(producer_edge.data.dst_subset)
                    for consumer_edge in consumer_edges
                ):
                    return False

            else:
                return False
        return True



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import ast
import collections.abc
import contextlib
import dataclasses
import inspect
import textwrap
import types
import typing
import warnings
from collections.abc import Callable
from typing import Generic, ParamSpec, TypeVar


P = ParamSpec("P")
T = TypeVar("T")


def _get_unique_name(func: Callable) -> str:
    """Generate a unique name for a callable object."""
    return (
        f"{func.__module__}.{getattr(func, '__qualname__', func.__class__.__qualname__)}#{id(func)}"
    )


def _is_empty_function(func: Callable) -> bool:
    """Check if a callable object is empty (i.e., contains no statements)."""
    try:
        callable_src = (
            inspect.getsource(func)
            if isinstance(func, types.FunctionType)
            else inspect.getsource(func.__call__)  # type: ignore[operator]  # asserted above
        )
        callable_ast = ast.parse(textwrap.dedent(callable_src))
        return all(
            isinstance(st, ast.Pass)
            or (isinstance(st, ast.Expr) and isinstance(st.value, ast.Constant))
            for st in typing.cast(ast.FunctionDef, callable_ast.body[0]).body
        )
    except Exception:
        return False


@dataclasses.dataclass(slots=True)
class _BaseHook(Generic[T, P]):
    """Base class to define callback registration functionality for all hook types."""

    definition: Callable[P, T]
    registry: dict[str, Callable[P, T]] = dataclasses.field(default_factory=dict, kw_only=True)
    callbacks: tuple[Callable[P, T], ...] = dataclasses.field(default=(), init=False)

    @property
    def __doc__(self) -> str | None:  # type: ignore[override]
        return self.definition.__doc__

    def __post_init__(self) -> None:
        # As an optimization to avoid an empty function call if no callbacks are
        # registered, we only add the original definitions to the list of callables
        # if it contains a non-empty definition.
        if not _is_empty_function(self.definition):
            self.callbacks = (self.definition,)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        raise NotImplementedError("This method should be implemented by subclasses.")

    def register(
        self, callback: Callable[P, T], *, name: str | None = None, index: int | None = None
    ) -> None:
        """
        Register a callback to the hook.

        Args:
            callback: The callable to register.
            name: An optional name for the callback. If not provided, a unique name will be generated.
            index: An optional index at which to insert the callback (not counting the original
               definition). If not provided, the callback will be appended to the end of the list.
        """

        callable_signature = inspect.signature(callback)
        hook_signature = inspect.signature(self.definition)

        signature_mismatch = len(callable_signature.parameters) != len(
            hook_signature.parameters
        ) or any(
            # Remove the annotation before comparison to avoid false mismatches
            actual_param.replace(annotation="") != expected_param.replace(annotation="")
            for actual_param, expected_param in zip(
                callable_signature.parameters.values(), hook_signature.parameters.values()
            )
        )
        if signature_mismatch:
            raise ValueError(
                f"Callback signature {callable_signature} does not match hook signature {hook_signature}"
            )
        try:
            callable_typing = typing.get_type_hints(callback)
            hook_typing = typing.get_type_hints(self.definition)
            if not all(
                callable_typing[arg_key] == arg_typing
                for arg_key, arg_typing in hook_typing.items()
            ):
                warnings.warn(
                    f"Callback annotations {callable_typing} does not match expected hook annotations {hook_typing}",
                    stacklevel=2,
                )
        except Exception:
            # Ignore issues while checking type hints (e.g., forward references
            # or missing imports); failure here should not prevent hook registration.
            pass

        name = name or _get_unique_name(callback)

        if index is None:
            self.callbacks += (callback,)
        else:
            if self.callbacks and self.callbacks[0] is self.definition:
                index += 1  # The original definition should always go first
            self.callbacks = (*self.callbacks[:index], callback, *self.callbacks[index:])

        self.registry[name] = callback

    def remove(self, callback: str | Callable[P, T]) -> None:
        """
        Remove a registered callback from the hook.

        Args:
            callback: The callable object to remove or its registered name.
        """
        if isinstance(callback, str):
            name = callback
            if name not in self.registry:
                raise KeyError(f"No callback registered under the name '{name}'")
        else:
            name = _get_unique_name(callback)
            if name not in self.registry:
                raise KeyError(f"Callback object {callback} not found in registry")

        callback = self.registry.pop(name)
        assert callback in self.callbacks
        self.callbacks = tuple(cb for cb in self.callbacks if cb is not callback)


@dataclasses.dataclass(slots=True)
class EventHook(_BaseHook[None, P]):
    """Event hook specification."""

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> None:
        for func in self.callbacks:
            func(*args, **kwargs)


def event_hook(definition: Callable[P, None]) -> EventHook[P]:
    """Decorator to create an EventHook from a function definition."""
    return EventHook(definition)


@dataclasses.dataclass(slots=True)
class ContextHook(
    contextlib.AbstractContextManager, _BaseHook[contextlib.AbstractContextManager, P]
):
    """
    Context hook specification.

    This hook type is used to define context managers that can be stacked together.
    """

    ctx_managers: collections.abc.Sequence[contextlib.AbstractContextManager] = dataclasses.field(
        default=(), init=False
    )

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> contextlib.AbstractContextManager:
        self.ctx_managers = [func(*args, **kwargs) for func in self.callbacks]
        return self

    def __enter__(self) -> None:
        for ctx_manager in self.ctx_managers:
            ctx_manager.__enter__()

    def __exit__(
        self,
        type_: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        for ctx_manager in reversed(self.ctx_managers):
            ctx_manager.__exit__(type_, exc_value, traceback)
        self.ctx_managers = ()


def context_hook(definition: Callable[P, contextlib.AbstractContextManager]) -> ContextHook[P]:
    """Decorator to create a ContextHook from a function definition."""
    return ContextHook(definition)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause
import dataclasses
from typing import TypeVar

from gt4py.eve import NodeTranslator, PreserveLocationVisitor
from gt4py.eve.extended_typing import Container, Self
from gt4py.next import common
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm, domain_utils


def _filter_domain(
    domain: domain_utils.SymbolicDomain, dims: Container[common.Dimension]
) -> domain_utils.SymbolicDomain:
    return domain_utils.SymbolicDomain(
        grid_type=domain.grid_type,
        ranges={d: r for d, r in domain.ranges.items() if d in dims},
    )


PRG = TypeVar("PRG", bound=itir.Program | itir.Expr)


@dataclasses.dataclass
class _PruneEmptyConcatWhere(PreserveLocationVisitor, NodeTranslator):
    """
    Prune `concat_where` expression with one branch never being accessed.

    This pass requires domain inference to be executed before.

    This pass the true and false branch values to be fields, not tuples of fields. Execute
     `gt4py.next.iterator.transforms.concat_where.expand_tuple_args` before.

    >>> from gt4py.next.iterator.ir_utils import ir_makers as im
    >>> IDim = common.Dimension("IDim")
    >>> expr = im.concat_where(im.domain(common.GridType.UNSTRUCTURED, {IDim: (0, 0)}), "a", "b")
    >>> assert prune_empty_concat_where(expr) == im.ref("b")
    """

    PRESERVED_ANNEX_ATTRS = ("domain",)

    @classmethod
    def apply(cls: type[Self], node: PRG) -> PRG:
        return cls().visit(node)

    def visit_FunCall(self, node: itir.FunCall) -> itir.Expr:
        node = self.generic_visit(node)

        if cpm.is_call_to(node, "concat_where"):
            cond_expr, tb, fb = node.args

            if tb == fb:
                # note: as long as we visited the args we have a copy here, so no need to copy again
                tb.annex.domain = node.annex.domain
                return tb

            cond = domain_utils.SymbolicDomain.from_expr(cond_expr)
            if cond.empty():
                return node.args[2]

            tb_domain, fb_domain = (
                _filter_domain(arg.annex.domain, cond.ranges.keys()) for arg in node.args[1:]
            )
            assert all(isinstance(d, domain_utils.SymbolicDomain) for d in (tb_domain, fb_domain))
            if tb_domain.empty():
                return node.args[2]
            if fb_domain.empty():
                return node.args[1]

        return node


prune_empty_concat_where = _PruneEmptyConcatWhere.apply



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
from typing import Sequence, TypedDict

from gt4py.next import common, config


class CompilationOptionsArgs(TypedDict, total=False):
    enable_jit: bool
    static_params: Sequence[str]
    connectivities: common.OffsetProvider
    static_domains: bool


@dataclasses.dataclass(frozen=True)
class CompilationOptions:
    #: Enable Just-in-Time compilation, otherwise a program has to be compiled manually by a call
    #: to `compile` before calling.
    # Uses a factory to make changes to the config after module import time take effect. This is
    # mostly important for testing. Users should not rely on it.
    enable_jit: bool = dataclasses.field(default_factory=lambda: config.ENABLE_JIT_DEFAULT)

    #: If the user requests static params, they will be used later to initialize CompiledPrograms.
    #: By default the set of static params is set when compiling for the first time, e.g. on call
    #: when jitting is enabled, or on a call to `compile`.
    static_params: Sequence[str] | None = None

    # TODO(ricoh): replace with common.OffsetProviderType once the temporary pass doesn't require the runtime information
    #: A dictionary holding static/compile-time information about the offset providers.
    #: For now, it is used for ahead of time compilation in DaCe orchestrated programs,
    #: i.e. DaCe programs that call GT4Py Programs -SDFGConvertible interface-.
    connectivities: common.OffsetProvider | None = None

    static_domains: bool = False


assert CompilationOptionsArgs.__annotations__.keys() == CompilationOptions.__annotations__.keys()



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
from collections.abc import Iterable, Iterator
from typing import TypeGuard

from gt4py.eve import NodeTranslator, PreserveLocationVisitor
from gt4py.next import common, utils
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm, ir_makers as im
from gt4py.next.iterator.ir_utils.common_pattern_matcher import is_applied_lift
from gt4py.next.type_system import type_specifications as ts


def _is_neighbors(arg: itir.Expr) -> TypeGuard[itir.FunCall]:
    return isinstance(arg, itir.FunCall) and arg.fun == itir.SymRef(id="neighbors")


def _is_neighbors_or_lifted_and_neighbors(arg: itir.Expr) -> TypeGuard[itir.FunCall]:
    return _is_neighbors(arg) or (
        is_applied_lift(arg)
        and any(_is_neighbors_or_lifted_and_neighbors(nested_arg) for nested_arg in arg.args)
    )


def _get_neighbors_args(reduce_args: Iterable[itir.Expr]) -> Iterator[itir.FunCall]:
    flat_reduce_args: list[itir.Expr] = []
    for arg in reduce_args:
        if cpm.is_call_to(arg, "if_"):
            flat_reduce_args.extend(_get_neighbors_args(arg.args[1:3]))
        else:
            flat_reduce_args.append(arg)

    return filter(_is_neighbors_or_lifted_and_neighbors, flat_reduce_args)


def _get_partial_offset_tags(reduce_args: Iterable[itir.Expr]) -> Iterable[str]:
    assert all(isinstance(arg.type, ts.ListType) for arg in reduce_args)

    return [
        arg.type.offset_type.value  # type: ignore[union-attr] # checked in previous lines
        for arg in reduce_args
        if arg.type.offset_type is not None  # type: ignore[union-attr] # checked in previous lines
    ]


def _get_connectivity(
    applied_reduce_node: itir.FunCall,
    offset_provider_type: common.OffsetProviderType,
) -> common.NeighborConnectivityType:
    """Return single connectivity that is compatible with the arguments of the reduce."""
    if not cpm.is_applied_reduce(applied_reduce_node):
        raise ValueError("Expected a call to a 'reduce' object, i.e. 'reduce(...)(...)'.")

    connectivities: list[common.NeighborConnectivityType] = []
    for o in _get_partial_offset_tags(applied_reduce_node.args):
        conn = common.get_offset_type(offset_provider_type, o)
        assert isinstance(conn, common.NeighborConnectivityType)
        connectivities.append(conn)

    if not connectivities:
        raise RuntimeError(
            "Couldn't deduce 'offset_type' in any arguments of 'reduce'. Did you reduce over `make_const_list`s only?"
        )

    if len({(c.max_neighbors, c.has_skip_values) for c in connectivities}) != 1:
        # The condition for this check is required but not sufficient: the actual neighbor tables could still be incompatible.
        raise RuntimeError("Arguments to 'reduce' have incompatible partial shifts.")
    return connectivities[0]


@dataclasses.dataclass(frozen=True)
class UnrollReduce(PreserveLocationVisitor, NodeTranslator):
    # we use one UID generator per instance such that the generated ids are
    # stable across multiple runs (required for caching to properly work)
    uids: utils.IDGeneratorPool = dataclasses.field(repr=False)

    @classmethod
    def apply(
        cls,
        node: itir.Node,
        offset_provider_type: common.OffsetProviderType,
        uids: utils.IDGeneratorPool,
    ) -> itir.Node:
        return cls(uids=uids).visit(node, offset_provider_type=offset_provider_type)

    def _visit_reduce(
        self, node: itir.FunCall, offset_provider_type: common.OffsetProviderType
    ) -> itir.Expr:
        connectivity_type = _get_connectivity(node, offset_provider_type)
        max_neighbors = connectivity_type.max_neighbors
        has_skip_values = connectivity_type.has_skip_values

        acc: str = next(self.uids["_acc"])
        offset: str = next(self.uids["_i"])
        step: str = next(self.uids["_step"])

        assert isinstance(node.fun, itir.FunCall)
        fun, init = node.fun.args

        elems = [im.list_get(offset, arg) for arg in node.args]
        step_fun: itir.Expr = im.call(fun)(acc, *elems)
        if has_skip_values:
            check_arg = next(_get_neighbors_args(node.args))
            offset_tag, it = check_arg.args
            can_deref = im.can_deref(im.shift(offset_tag, offset)(it))
            step_fun = im.if_(can_deref, step_fun, acc)
        step_fun = im.lambda_(acc, offset)(step_fun)
        expr = init
        for i in range(max_neighbors):
            expr = im.call(step)(expr, itir.OffsetLiteral(value=i))
        expr = im.let(step, step_fun)(expr)

        return expr

    def visit_FunCall(self, node: itir.FunCall, **kwargs) -> itir.Expr:
        node = self.generic_visit(node, **kwargs)
        if cpm.is_applied_reduce(node):
            return self._visit_reduce(node, **kwargs)
        return node



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
import functools
from typing import Any, Callable, ClassVar, Iterable, Optional, Type, TypeGuard, Union

import gt4py.eve as eve
from gt4py.eve.concepts import SymbolName
from gt4py.next import common, utils
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import (
    common_pattern_matcher as cpm,
    ir_makers as im,
    misc as ir_utils_misc,
)
from gt4py.next.iterator.type_system import inference as itir_type_inference
from gt4py.next.otf import cpp_utils
from gt4py.next.program_processors.codegens.gtfn.gtfn_ir import (
    Backend,
    BinaryExpr,
    CartesianDomain,
    CastExpr,
    FunCall,
    FunctionDefinition,
    IfStmt,
    IntegralConstant,
    Lambda,
    Literal,
    OffsetLiteral,
    Program,
    Scan,
    ScanExecution,
    ScanPassDefinition,
    SidComposite,
    SidFromScalar,
    StencilExecution,
    TagDefinition,
    TaggedValues,
    TemporaryAllocation,
    TernaryExpr,
    UnaryExpr,
    UnstructuredDomain,
)
from gt4py.next.program_processors.codegens.gtfn.gtfn_ir_common import Expr, Node, Sym, SymRef
from gt4py.next.type_system import type_info, type_specifications as ts


_vertical_dimension = "gtfn::unstructured::dim::vertical"
_horizontal_dimension = "gtfn::unstructured::dim::horizontal"


def _is_tuple_of_ref_or_literal(expr: itir.Expr) -> bool:
    if (
        isinstance(expr, itir.FunCall)
        and isinstance(expr.fun, itir.SymRef)
        and expr.fun.id == "tuple_get"
        and len(expr.args) == 2
        and _is_tuple_of_ref_or_literal(expr.args[1])
    ):
        return True
    if (
        isinstance(expr, itir.FunCall)
        and isinstance(expr.fun, itir.SymRef)
        and expr.fun.id == "make_tuple"
        and all(_is_tuple_of_ref_or_literal(arg) for arg in expr.args)
    ):
        return True
    if isinstance(expr, (itir.SymRef, itir.Literal)):
        return True
    return False


def _get_domains(nodes: Iterable[itir.Stmt]) -> Iterable[itir.FunCall]:
    result = set()
    for node in nodes:
        result |= node.walk_values().if_isinstance(itir.SetAt).getattr("domain").to_set()
    return result


def _name_from_named_range(named_range_call: itir.FunCall) -> str:
    assert isinstance(named_range_call, itir.FunCall) and named_range_call.fun == itir.SymRef(
        id="named_range"
    )
    assert isinstance(named_range_call.args[0], itir.AxisLiteral)
    return named_range_call.args[0].value


def _collect_dimensions_from_domain(
    body: Iterable[itir.Stmt],
) -> dict[str, TagDefinition]:
    domains = _get_domains(body)
    offset_definitions = {}
    for domain in domains:
        if domain.fun == itir.SymRef(id="cartesian_domain"):
            for nr in domain.args:
                assert isinstance(nr, itir.FunCall)
                dim_name = _name_from_named_range(nr)
                offset_definitions[dim_name] = TagDefinition(name=Sym(id=dim_name))
        elif domain.fun == itir.SymRef(id="unstructured_domain"):
            if len(domain.args) > 2:
                raise ValueError("Unstructured_domain must not have more than 2 arguments.")
            if len(domain.args) > 0:
                horizontal_range = domain.args[0]
                assert isinstance(horizontal_range, itir.FunCall)
                horizontal_name = _name_from_named_range(horizontal_range)
                offset_definitions[horizontal_name] = TagDefinition(
                    name=Sym(id=horizontal_name), alias=_horizontal_dimension
                )
            if len(domain.args) > 1:
                vertical_range = domain.args[1]
                assert isinstance(vertical_range, itir.FunCall)
                vertical_name = _name_from_named_range(vertical_range)
                offset_definitions[vertical_name] = TagDefinition(
                    name=Sym(id=vertical_name), alias=_vertical_dimension
                )
        else:
            raise AssertionError(
                "Expected either a call to 'cartesian_domain' or to 'unstructured_domain'."
            )
    return offset_definitions


def _collect_offset_definitions(
    node: itir.Node,
    grid_type: common.GridType,
    offset_provider_type: common.OffsetProviderType,
) -> dict[str, TagDefinition]:
    used_offset_tags: set[str] = (
        node.walk_values()
        .if_isinstance(itir.OffsetLiteral)
        .filter(lambda offset_literal: isinstance(offset_literal.value, str))
        .getattr("value")
    ).to_set()
    # implicit offsets don't occur in the `offset_provider_type`, get them from the used offset tags
    offset_provider_type = {
        offset_name: common.get_offset_type(offset_provider_type, offset_name)
        for offset_name in used_offset_tags
    } | {**offset_provider_type}
    offset_definitions = {}

    for offset_name, dim_or_connectivity_type in offset_provider_type.items():
        if isinstance(dim_or_connectivity_type, common.Dimension):
            dim: common.Dimension = dim_or_connectivity_type
            if grid_type == common.GridType.CARTESIAN:
                # create alias from offset to dimension
                offset_definitions[dim.value] = TagDefinition(name=Sym(id=dim.value))
                offset_definitions[offset_name] = TagDefinition(
                    name=Sym(id=offset_name), alias=SymRef(id=dim.value)
                )
            else:
                assert grid_type == common.GridType.UNSTRUCTURED
                if not dim.kind == common.DimensionKind.VERTICAL:
                    raise ValueError(
                        "Mapping an offset to a horizontal dimension in unstructured is not allowed."
                    )
                # create alias from vertical offset to vertical dimension
                offset_definitions[dim.value] = TagDefinition(
                    name=Sym(id=dim.value), alias=_vertical_dimension
                )
                offset_definitions[offset_name] = TagDefinition(
                    name=Sym(id=offset_name), alias=SymRef(id=dim.value)
                )
        elif isinstance(
            connectivity_type := dim_or_connectivity_type, common.NeighborConnectivityType
        ):
            assert grid_type == common.GridType.UNSTRUCTURED
            offset_definitions[offset_name] = TagDefinition(name=Sym(id=offset_name))
            if offset_name != connectivity_type.neighbor_dim.value:
                offset_definitions[connectivity_type.neighbor_dim.value] = TagDefinition(
                    name=Sym(id=connectivity_type.neighbor_dim.value)
                )

            for dim in [connectivity_type.source_dim, connectivity_type.codomain]:
                if dim.kind != common.DimensionKind.HORIZONTAL:
                    raise NotImplementedError()
                offset_definitions[dim.value] = TagDefinition(
                    name=Sym(id=dim.value), alias=_horizontal_dimension
                )
        else:
            raise AssertionError(
                "Elements of offset provider need to be either 'Dimension' or 'Connectivity'."
            )
    return offset_definitions


def _literal_as_integral_constant(node: itir.Literal) -> IntegralConstant:
    assert type_info.is_integer(node.type)
    return IntegralConstant(value=int(node.value))


def _is_scan(node: itir.Node) -> TypeGuard[itir.FunCall]:
    return isinstance(node, itir.FunCall) and node.fun == itir.SymRef(id="scan")


def _bool_from_literal(node: itir.Node) -> bool:
    assert isinstance(node, itir.Literal)
    assert type_info.is_logical(node.type) and node.value in ("True", "False")
    return node.value == "True"


class _CannonicalizeUnstructuredDomain(eve.NodeTranslator):
    def visit_FunCall(self, node: itir.FunCall) -> itir.FunCall:
        if node.fun == itir.SymRef(id="unstructured_domain"):
            # for no good reason, the domain arguments for unstructured need to be in order (horizontal, vertical)
            assert isinstance(node.args[0], itir.FunCall)
            first_axis_literal = node.args[0].args[0]
            assert isinstance(first_axis_literal, itir.AxisLiteral)
            if first_axis_literal.kind == itir.DimensionKind.VERTICAL:
                assert len(node.args) == 2
                assert isinstance(node.args[1], itir.FunCall)
                assert isinstance(node.args[1].args[0], itir.AxisLiteral)
                assert node.args[1].args[0].kind == itir.DimensionKind.HORIZONTAL
                return itir.FunCall(fun=node.fun, args=[node.args[1], node.args[0]])
        return node

    @classmethod
    def apply(
        cls,
        node: itir.Program,
    ) -> itir.Program:
        if not isinstance(node, itir.Program):
            raise TypeError(f"Expected a 'Program', got '{type(node).__name__}'.")

        return cls().visit(node)


def _process_elements(
    process_func: Callable[..., Expr],
    obj: Expr,
    type_: ts.TypeSpec,
    *,
    tuple_constructor: Callable[..., Expr] = lambda _, *elements: FunCall(
        fun=SymRef(id="make_tuple"), args=list(elements)
    ),
) -> Expr:
    """
    Recursively applies a processing function to all primitive constituents of a tuple.

    Be aware that this function duplicates the `obj` expression and should hence be used with care.

    Arguments:
        process_func: A callable that takes a gtfn_ir.Expr representing a leaf-element of `obj`.
        obj: The object whose elements are to be transformed.
        type_: A type with the same structure as the elements of `obj`.
        tuple_constructor: By default all transformed tuple elements are just put in a tuple again.
            This can be customized by passing a different Callable.
    """
    assert isinstance(type_, ts.TypeSpec)

    def _gen_constituent_expr(el_type: ts.ScalarType | ts.FieldType, path: tuple[int, ...]) -> Expr:
        # construct expression for the currently processed element
        el = functools.reduce(
            lambda cur_expr, i: FunCall(
                fun=SymRef(id="tuple_get"), args=[IntegralConstant(value=i), cur_expr]
            ),
            path,
            obj,
        )
        return process_func(el, el_type)

    result = type_info.apply_to_primitive_constituents(
        _gen_constituent_expr,
        type_,
        with_path_arg=True,
        tuple_constructor=tuple_constructor,
    )
    return result


@dataclasses.dataclass(frozen=True)
class GTFN_lowering(eve.NodeTranslator, eve.VisitorWithSymbolTableTrait):
    _binary_op_map: ClassVar[dict[str, str]] = {
        "plus": "+",
        "minus": "-",
        "multiplies": "*",
        "divides": "/",
        "eq": "==",
        "not_eq": "!=",
        "less": "<",
        "less_equal": "<=",
        "greater": ">",
        "greater_equal": ">=",
        "and_": "&&",
        "or_": "||",
        "xor_": "^",
        "mod": "%",
    }
    _unary_op_map: ClassVar[dict[str, str]] = {"not_": "!"}

    offset_provider_type: common.OffsetProviderType
    column_axis: Optional[common.Dimension]
    grid_type: common.GridType

    # we use one UID generator per instance such that the generated ids are
    # stable across multiple runs (required for caching to properly work)
    uids: utils.IDGeneratorPool = dataclasses.field(
        init=False, repr=False, default_factory=utils.IDGeneratorPool
    )

    @classmethod
    def apply(
        cls,
        node: itir.Program,
        *,
        offset_provider_type: common.OffsetProviderType,
        column_axis: Optional[common.Dimension],
    ) -> Program:
        if not isinstance(node, itir.Program):
            raise TypeError(f"Expected a 'Program', got '{type(node).__name__}'.")

        node = itir_type_inference.infer(node, offset_provider_type=offset_provider_type)
        grid_type = ir_utils_misc.grid_type_from_program(node)
        if grid_type == common.GridType.UNSTRUCTURED:
            node = _CannonicalizeUnstructuredDomain.apply(node)
        return cls(
            offset_provider_type=offset_provider_type, column_axis=column_axis, grid_type=grid_type
        ).visit(node)

    def visit_Sym(self, node: itir.Sym, **kwargs: Any) -> Sym:
        return Sym(id=node.id)

    def visit_SymRef(
        self,
        node: itir.SymRef,
        force_function_extraction: bool = False,
        extracted_functions: Optional[list] = None,
        **kwargs: Any,
    ) -> SymRef:
        if force_function_extraction and node.id == "deref":
            assert extracted_functions is not None
            fun_id = next(self.uids["_fun"])
            fun_def = FunctionDefinition(
                id=fun_id,
                params=[Sym(id="x")],
                expr=FunCall(fun=SymRef(id="deref"), args=[SymRef(id="x")]),
            )
            extracted_functions.append(fun_def)
            return SymRef(id=fun_id)
        return SymRef(id=node.id)

    def visit_Lambda(
        self,
        node: itir.Lambda,
        *,
        force_function_extraction: bool = False,
        extracted_functions: Optional[list] = None,
        **kwargs: Any,
    ) -> Union[SymRef, Lambda]:
        if force_function_extraction:
            assert extracted_functions is not None
            fun_id = next(self.uids["_fun"])
            fun_def = FunctionDefinition(
                id=fun_id,
                params=self.visit(node.params, **kwargs),
                expr=self.visit(node.expr, **kwargs),
            )
            extracted_functions.append(fun_def)
            return SymRef(id=fun_id)
        return Lambda(
            params=self.visit(node.params, **kwargs), expr=self.visit(node.expr, **kwargs)
        )

    def visit_Literal(self, node: itir.Literal, **kwargs: Any) -> Literal:
        return Literal(value=node.value, type=node.type.kind.name.lower())

    def visit_OffsetLiteral(self, node: itir.OffsetLiteral, **kwargs: Any) -> OffsetLiteral:
        return OffsetLiteral(value=node.value)

    def visit_AxisLiteral(self, node: itir.AxisLiteral, **kwargs: Any) -> Literal:
        return Literal(value=node.value, type="axis_literal")

    def _make_domain(self, node: itir.FunCall) -> tuple[TaggedValues, TaggedValues]:
        tags = []
        sizes = []
        offsets = []
        for named_range in node.args:
            if not (
                isinstance(named_range, itir.FunCall)
                and named_range.fun == itir.SymRef(id="named_range")
            ):
                raise ValueError("Arguments to 'domain' need to be calls to 'named_range'.")
            tags.append(self.visit(named_range.args[0]))
            sizes.append(
                BinaryExpr(
                    op="-", lhs=self.visit(named_range.args[2]), rhs=self.visit(named_range.args[1])
                )
            )
            offsets.append(self.visit(named_range.args[1]))
        return TaggedValues(tags=tags, values=sizes), TaggedValues(tags=tags, values=offsets)

    @staticmethod
    def _collect_offset_or_axis_node(
        node_type: Type, tree: eve.Node | Iterable[eve.Node]
    ) -> set[str]:
        if not isinstance(tree, Iterable):
            tree = [tree]
        result = set()
        for n in tree:
            result.update(
                n.pre_walk_values()
                .if_isinstance(node_type)
                .getattr("value")
                .if_isinstance(str)
                .to_set()
            )
        return result

    def _visit_if_(self, node: itir.FunCall, **kwargs: Any) -> Node:
        assert len(node.args) == 3
        return TernaryExpr(
            cond=self.visit(node.args[0], **kwargs),
            true_expr=self.visit(node.args[1], **kwargs),
            false_expr=self.visit(node.args[2], **kwargs),
        )

    def _visit_cast_(self, node: itir.FunCall, **kwargs: Any) -> Node:
        assert len(node.args) == 2
        return CastExpr(
            obj_expr=self.visit(node.args[0], **kwargs),
            new_dtype=self.visit(node.args[1], **kwargs),
        )

    def _visit_tuple_get(self, node: itir.FunCall, **kwargs: Any) -> Node:
        assert isinstance(node.args[0], itir.Literal)
        return FunCall(
            fun=SymRef(id="tuple_get"),
            args=[_literal_as_integral_constant(node.args[0]), self.visit(node.args[1])],
        )

    def _visit_list_get(self, node: itir.FunCall, **kwargs: Any) -> Node:
        # should only reach this for the case of an external sparse field
        tuple_idx = (
            _literal_as_integral_constant(node.args[0])
            if isinstance(node.args[0], itir.Literal)
            else self.visit(
                node.args[0]
            )  # from unroll_reduce we get a `SymRef` which is refering to an `OffsetLiteral` which is lowered to integral_constant
        )
        return FunCall(fun=SymRef(id="tuple_get"), args=[tuple_idx, self.visit(node.args[1])])

    def _visit_cartesian_domain(self, node: itir.FunCall, **kwargs: Any) -> Node:
        sizes, domain_offsets = self._make_domain(node)
        return CartesianDomain(tagged_sizes=sizes, tagged_offsets=domain_offsets)

    def _visit_unstructured_domain(self, node: itir.FunCall, **kwargs: Any) -> Node:
        sizes, domain_offsets = self._make_domain(node)
        connectivities = []
        if "stencil" in kwargs:
            shift_offsets = self._collect_offset_or_axis_node(itir.OffsetLiteral, kwargs["stencil"])
            for o in shift_offsets:
                if o in self.offset_provider_type and isinstance(
                    common.get_offset_type(self.offset_provider_type, o),
                    common.NeighborConnectivityType,
                ):
                    connectivities.append(SymRef(id=o))
        return UnstructuredDomain(
            tagged_sizes=sizes, tagged_offsets=domain_offsets, connectivities=connectivities
        )

    def _visit_get_domain_range(self, node: itir.FunCall, **kwargs: Any) -> Node:
        field, dim = node.args

        return FunCall(
            fun=SymRef(id="get_domain_range"),
            args=[self.visit(field, **kwargs), self.visit(dim, **kwargs)],
        )

    def visit_FunCall(self, node: itir.FunCall, **kwargs: Any) -> Node:
        if isinstance(node.fun, itir.SymRef):
            if node.fun.id in self._unary_op_map:
                assert len(node.args) == 1
                return UnaryExpr(
                    op=self._unary_op_map[node.fun.id], expr=self.visit(node.args[0], **kwargs)
                )
            elif node.fun.id in self._binary_op_map:
                assert len(node.args) == 2
                return BinaryExpr(
                    op=self._binary_op_map[node.fun.id],
                    lhs=self.visit(node.args[0], **kwargs),
                    rhs=self.visit(node.args[1], **kwargs),
                )
            elif hasattr(self, visit_method := f"_visit_{node.fun.id}"):
                # special handling of applied builtins is handled in `_visit_<builtin>`
                return getattr(self, visit_method)(node, **kwargs)
            elif node.fun.id == "shift":
                raise ValueError("Unapplied shift call not supported: '{node}'.")
            elif node.fun.id == "scan":
                raise ValueError("Scans are only supported at the top level of a stencil closure.")
        if isinstance(node.fun, itir.FunCall):
            if node.fun.fun == itir.SymRef(id="shift"):
                assert len(node.args) == 1
                return FunCall(
                    fun=self.visit(node.fun.fun, **kwargs),
                    args=self.visit(node.args, **kwargs) + self.visit(node.fun.args, **kwargs),
                )
        return FunCall(fun=self.visit(node.fun, **kwargs), args=self.visit(node.args, **kwargs))

    def visit_FunctionDefinition(
        self, node: itir.FunctionDefinition, **kwargs: Any
    ) -> FunctionDefinition:
        return FunctionDefinition(
            id=node.id,
            params=self.visit(node.params, **kwargs),
            expr=self.visit(node.expr, **kwargs),
        )

    def _visit_output_argument(self, node: itir.Expr) -> SidComposite | SymRef:
        lowered_output = self.visit(node)

        # just a sanity check, identity function otherwise
        def check_el_type(el_expr: Expr, el_type: ts.ScalarType | ts.FieldType) -> Expr:
            assert isinstance(el_type, ts.FieldType)
            return el_expr

        assert isinstance(node.type, ts.TypeSpec)
        lowered_output_as_sid = _process_elements(
            check_el_type,
            lowered_output,
            node.type,
            tuple_constructor=lambda *elements: SidComposite(values=list(elements)),
        )

        assert isinstance(lowered_output_as_sid, (SidComposite, SymRef))
        return lowered_output_as_sid

    @staticmethod
    def _merge_scans(
        executions: list[Union[StencilExecution, ScanExecution]],
    ) -> list[Union[StencilExecution, ScanExecution]]:
        def merge(a: ScanExecution, b: ScanExecution) -> ScanExecution:
            assert a.backend == b.backend
            assert a.axis == b.axis

            index_map = dict[int, int]()
            compacted_b_args = list[Expr]()
            for b_idx, b_arg in enumerate(b.args):
                try:
                    a_idx = a.args.index(b_arg)
                    index_map[b_idx] = a_idx
                except ValueError:
                    index_map[b_idx] = len(a.args) + len(compacted_b_args)
                    compacted_b_args.append(b_arg)

            def remap_args(s: Scan) -> Scan:
                return Scan(
                    function=s.function,
                    output=index_map[s.output],
                    inputs=[index_map[i] for i in s.inputs],
                    init=s.init,
                )

            return ScanExecution(
                backend=a.backend,
                scans=a.scans + [remap_args(s) for s in b.scans],
                args=a.args + compacted_b_args,
                axis=a.axis,
            )

        res = executions[:1]
        for execution in executions[1:]:
            if (
                isinstance(execution, ScanExecution)
                and isinstance(res[-1], ScanExecution)
                and execution.backend == res[-1].backend
            ):
                res[-1] = merge(res[-1], execution)
            else:
                res.append(execution)
        return res

    def visit_Stmt(self, node: itir.Stmt, **kwargs: Any) -> None:
        raise AssertionError("All Stmts need to be handled explicitly.")

    def visit_IfStmt(self, node: itir.IfStmt, **kwargs: Any) -> IfStmt:
        return IfStmt(
            cond=self.visit(node.cond, **kwargs),
            true_branch=self.visit(node.true_branch, **kwargs),
            false_branch=self.visit(node.false_branch, **kwargs),
        )

    def visit_SetAt(
        self, node: itir.SetAt, *, extracted_functions: list, **kwargs: Any
    ) -> Union[StencilExecution, ScanExecution]:
        if _is_tuple_of_ref_or_literal(node.expr):
            node.expr = im.as_fieldop("deref", node.domain)(node.expr)

        itir_projector, extracted_expr = ir_utils_misc.extract_projector(node.expr)
        projector = self.visit(itir_projector, **kwargs) if itir_projector is not None else None
        node.expr = extracted_expr

        assert cpm.is_applied_as_fieldop(node.expr), node.expr
        stencil = node.expr.fun.args[0]
        domain = node.domain
        inputs = node.expr.args
        lowered_inputs = []
        for input_ in inputs:
            lowered_input = self.visit(input_, **kwargs)

            # convert scalar elements into SIDs, leave rest as is
            def convert_el_to_sid(el_expr: Expr, el_type: ts.ScalarType | ts.FieldType) -> Expr:
                if isinstance(el_type, ts.ScalarType):
                    return SidFromScalar(arg=el_expr)
                else:
                    assert isinstance(el_type, ts.FieldType)
                    return el_expr

            assert isinstance(input_.type, ts.TypeSpec)
            lowered_input_as_sid = _process_elements(
                convert_el_to_sid,
                lowered_input,
                input_.type,
                tuple_constructor=lambda *elements: SidComposite(values=list(elements)),
            )

            lowered_inputs.append(lowered_input_as_sid)

        backend = Backend(domain=self.visit(domain, stencil=stencil, **kwargs))
        if _is_scan(stencil):
            scan_id = next(self.uids["_scan"])
            scan_lambda = self.visit(stencil.args[0], **kwargs)
            forward = _bool_from_literal(stencil.args[1])
            scan_def = ScanPassDefinition(
                id=scan_id,
                params=scan_lambda.params,
                expr=scan_lambda.expr,
                forward=forward,
                projector=projector,
            )
            extracted_functions.append(scan_def)
            scan = Scan(
                function=SymRef(id=scan_id),
                output=0,
                inputs=[i + 1 for i, _ in enumerate(inputs)],
                init=self.visit(stencil.args[2], **kwargs),
            )
            column_axis = self.column_axis
            assert isinstance(column_axis, common.Dimension)
            return ScanExecution(
                backend=backend,
                scans=[scan],
                args=[self._visit_output_argument(node.target), *lowered_inputs],
                axis=SymRef(id=column_axis.value),
            )
        assert projector is None  # only scans have projectors
        return StencilExecution(
            stencil=self.visit(
                stencil,
                force_function_extraction=True,
                extracted_functions=extracted_functions,
                **kwargs,
            ),
            output=self._visit_output_argument(node.target),
            inputs=lowered_inputs,
            backend=backend,
        )

    def visit_Program(self, node: itir.Program, **kwargs: Any) -> Program:
        extracted_functions: list[Union[FunctionDefinition, ScanPassDefinition]] = []
        executions = self.visit(node.body, extracted_functions=extracted_functions)
        executions = self._merge_scans(executions)
        function_definitions = self.visit(node.function_definitions) + extracted_functions
        offset_definitions = {
            **_collect_dimensions_from_domain(node.body),
            **_collect_offset_definitions(node, self.grid_type, self.offset_provider_type),
        }
        return Program(
            id=SymbolName(node.id),
            params=self.visit(node.params),
            executions=executions,
            grid_type=self.grid_type,
            offset_definitions=list(offset_definitions.values()),
            function_definitions=function_definitions,
            temporaries=self.visit(node.declarations, params=[p.id for p in node.params]),
        )

    def visit_Temporary(
        self, node: itir.Temporary, *, params: list, **kwargs: Any
    ) -> TemporaryAllocation:
        def dtype_to_cpp(x: ts.DataType) -> str:
            if isinstance(x, ts.TupleType):
                assert all(isinstance(i, ts.ScalarType) for i in x.types)
                return "::gridtools::tuple<" + ", ".join(dtype_to_cpp(i) for i in x.types) + ">"  # type: ignore[arg-type] # ensured by assert
            assert isinstance(x, ts.ScalarType)
            res = cpp_utils.pytype_to_cpptype(x)
            assert isinstance(res, str)
            return res

        assert node.dtype
        return TemporaryAllocation(
            id=node.id, dtype=dtype_to_cpp(node.dtype), domain=self.visit(node.domain, **kwargs)
        )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import contextlib
import copy
import io
import os
import shutil
import threading
from typing import Any, Dict, List, Literal, Optional, Tuple, Type, TypedDict, Union

import pybind11
import setuptools
from setuptools import distutils
from setuptools.command.build_ext import build_ext

from gt4py._core import definitions as core_defs
from gt4py.cartesian import config as gt_config


_SETUPTOOLS_LOCK = threading.Lock()


class SetuptoolsArgs(TypedDict):
    name: str
    ext_modules: list[setuptools.Extension]
    script_args: list[str]


def get_dace_module_path() -> Optional[str]:
    try:
        import dace

        return os.path.dirname(dace.__file__)
    except ImportError:
        return None


def get_cuda_compute_capability():
    try:
        import cupy as cp

        return cp.cuda.Device(0).compute_capability
    except ImportError:
        return None


def get_gt_pyext_build_opts(
    *,
    debug_mode: bool = False,
    opt_level: Literal["0", "1", "2", "3", "s"] = "3",
    extra_opt_flags: str = "",
    add_profile_info: bool = False,
    uses_cuda: bool = False,
) -> Dict[str, Union[str, List[str], Dict[str, Any]]]:
    include_dirs: list[str] = []
    extra_compile_args_from_config = gt_config.build_settings["extra_compile_args"]
    is_rocm_gpu = core_defs.CUPY_DEVICE_TYPE == core_defs.DeviceType.ROCM

    if uses_cuda:
        if is_rocm_gpu:
            cuda_arch = gt_config.build_settings["cuda_arch"]
        else:
            compute_capability = get_cuda_compute_capability()
            cuda_arch = gt_config.build_settings["cuda_arch"] or compute_capability
            if not cuda_arch:
                raise RuntimeError("CUDA architecture could not be determined")
            if cuda_arch.startswith("sm_"):
                cuda_arch = cuda_arch.replace("sm_", "")
            if compute_capability and int(compute_capability) < int(cuda_arch):
                raise RuntimeError(
                    f"CUDA architecture {cuda_arch} exceeds compute capability {compute_capability}"
                )
    else:
        cuda_arch = ""

    gt_include_path = gt_config.build_settings["gt_include_path"]

    extra_compile_args = dict(
        cxx=[
            "-std=c++17",
            "-ftemplate-depth={}".format(gt_config.build_settings["cpp_template_depth"]),
            "-fvisibility=hidden",
            "-fPIC",
            # A compiler is allowed to choose if `char` is signed or unsigned. We force the signed behavior
            # because `char` is used to represent the `int8` type in GT4Py programs.
            "-fsigned-char",
            "-isystem{}".format(gt_include_path),
            *extra_compile_args_from_config["cxx"],
        ]
    )
    extra_compile_args["cuda"] = [
        "-std=c++17",
        "-ftemplate-depth={}".format(gt_config.build_settings["cpp_template_depth"]),
        *extra_compile_args_from_config["cuda"],
    ]
    if is_rocm_gpu:
        extra_compile_args["cuda"] += [
            "-isystem{}".format(gt_include_path),
            "-fvisibility=hidden",
            "-fPIC",
            *([f"--offload-arch={cuda_arch}"] if cuda_arch else []),
        ]
    else:
        extra_compile_args["cuda"] += [
            "-isystem={}".format(gt_include_path),
            "-arch=sm_{}".format(cuda_arch),
            "--expt-relaxed-constexpr",
            "--compiler-options",
            "-fvisibility=hidden",
            "--compiler-options",
            "-fPIC",
        ]
    extra_link_args = copy.deepcopy(gt_config.build_settings["extra_link_args"])

    mode_flags = (
        ["-O0", "-ggdb"] if debug_mode else [f"-O{opt_level}", "-DNDEBUG", *extra_opt_flags.split()]
    )

    extra_compile_args["cxx"].extend(mode_flags)
    extra_compile_args["cuda"].extend(mode_flags)
    extra_link_args.extend(mode_flags)

    if dace_path := get_dace_module_path():
        extra_compile_args["cxx"].append(
            "-isystem{}".format(os.path.join(dace_path, "runtime/include"))
        )
        if is_rocm_gpu:
            extra_compile_args["cuda"].append(
                "-isystem{}".format(os.path.join(dace_path, "runtime/include"))
            )
        else:
            extra_compile_args["cuda"].append(
                "-isystem={}".format(os.path.join(dace_path, "runtime/include"))
            )

    if add_profile_info:
        profile_flags = ["-pg"]
        extra_compile_args["cxx"].extend(profile_flags)
        extra_compile_args["cuda"].extend(profile_flags)
        extra_link_args.extend(profile_flags)

    if uses_cuda:
        build_opts = dict(
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args,
            extra_link_args=extra_link_args,
        )
    else:
        build_opts = dict(
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args["cxx"],
            extra_link_args=extra_link_args,
        )

    if gt_config.build_settings["openmp"]["use_openmp"]:
        cpp_flags = gt_config.build_settings["openmp"]["cppflags"]
        if uses_cuda:
            cuda_flags = []
            for cpp_flag in cpp_flags:
                if is_rocm_gpu:
                    cuda_flags.extend([cpp_flag])
                else:
                    cuda_flags.extend(["--compiler-options", cpp_flag])
            build_opts["extra_compile_args"]["cuda"].extend(cuda_flags)
        elif cpp_flags:
            build_opts["extra_compile_args"].extend(cpp_flags)

        ld_flags = gt_config.build_settings["openmp"]["ldflags"]
        if ld_flags:
            build_opts["extra_link_args"].extend(ld_flags)

    return build_opts


def setuptools_setup(*, build_ext_class: type[build_ext] | None, **kwargs) -> None:
    """
    Calls setuptools.setup() with 'cmdclass' set to 'build_ext_class'.

    This is a workaround because any config file that sets an element of
    'cmdclass' will override (instead of extend) the 'cmdclass' dict passed
    as argument to 'setuptools.setup()'.
    """
    with _SETUPTOOLS_LOCK:
        old_setup_stop_after = setuptools.distutils.core._setup_stop_after
        setuptools.distutils.core._setup_stop_after = "commandline"
        dist = setuptools.setup(**kwargs)
        if build_ext_class is not None:
            dist.cmdclass.update({"build_ext": build_ext_class})
        setuptools.distutils.core._setup_stop_after = old_setup_stop_after
    setuptools.distutils.core.run_commands(dist)


def build_pybind_ext(
    name: str,
    sources: list,
    build_path: str,
    target_path: str,
    *,
    include_dirs: Optional[List[str]] = None,
    library_dirs: Optional[List[str]] = None,
    libraries: Optional[List[str]] = None,
    extra_compile_args: Optional[Union[List[str], Dict[str, List[str]]]] = None,
    extra_link_args: Optional[List[str]] = None,
    build_ext_class: Optional[Type] = None,
    verbose: bool = False,
    clean: bool = False,
) -> Tuple[str, str]:
    # Hack to remove warning about "-Wstrict-prototypes" not having effect in C++
    replaced_flags_backup = copy.deepcopy(distutils.sysconfig._config_vars)
    _clean_build_flags(distutils.sysconfig._config_vars)

    include_dirs = include_dirs or []
    library_dirs = library_dirs or []
    libraries = libraries or []
    extra_compile_args = extra_compile_args or []
    extra_link_args = extra_link_args or []

    # Build extension module
    py_extension = setuptools.Extension(
        name,
        sources,
        include_dirs=[
            pybind11.get_include(),
            pybind11.get_include(user=True),
            *include_dirs,
        ],
        library_dirs=[*library_dirs],
        libraries=[*libraries],
        language="c++",
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )

    setuptools_args = SetuptoolsArgs(
        name=name,
        ext_modules=[py_extension],
        script_args=[
            "build_ext",
            "--build-temp={}".format(build_path),
            "--build-lib={}".format(build_path),
            "--force",
        ],
    )

    if verbose:
        setuptools_args["script_args"].append("-v")
        setuptools_setup(**setuptools_args, build_ext_class=build_ext_class)
    else:
        setuptools_args["script_args"].append("-q")
        io_out, io_err = io.StringIO(), io.StringIO()
        with contextlib.redirect_stdout(io_out), contextlib.redirect_stderr(io_err):
            setuptools_setup(**setuptools_args, build_ext_class=build_ext_class)

    # Copy extension in target path
    module_name = py_extension._full_name
    file_path = py_extension._file_name
    src_path = os.path.join(build_path, file_path)
    dest_path = os.path.join(target_path, os.path.basename(file_path))
    os.makedirs(os.path.dirname(dest_path), exist_ok=True)
    distutils.file_util.copy_file(src_path, dest_path, verbose=verbose)

    # Final cleaning
    if clean:
        shutil.rmtree(build_path)

    # Restore original distutils flag config to not break functionality with "-Wstrict-prototypes"-hack for other
    # tools using distutils.
    for key, value in replaced_flags_backup.items():
        distutils.sysconfig._config_vars[key] = value

    return module_name, dest_path


def build_pybind_cuda_ext(
    name: str,
    sources: list,
    build_path: str,
    target_path: str,
    *,
    include_dirs: Optional[List[str]] = None,
    library_dirs: Optional[List[str]] = None,
    libraries: Optional[List[str]] = None,
    extra_compile_args: Optional[Union[List[str], Dict[str, List[str]]]] = None,
    extra_link_args: Optional[List[str]] = None,
    verbose: bool = False,
    clean: bool = False,
) -> Tuple[str, str]:
    include_dirs = include_dirs or []
    include_dirs = [*include_dirs, gt_config.build_settings["cuda_include_path"]]
    library_dirs = library_dirs or []
    library_dirs = [*library_dirs, gt_config.build_settings["cuda_library_path"]]
    libraries = libraries or []
    if core_defs.CUPY_DEVICE_TYPE == core_defs.DeviceType.ROCM:
        libraries = [*libraries, "hiprtc"]
    else:
        libraries = [*libraries, "cudart"]
    extra_compile_args = extra_compile_args or []

    return build_pybind_ext(
        name,
        sources,
        build_path,
        target_path,
        verbose=verbose,
        clean=clean,
        include_dirs=include_dirs,
        library_dirs=library_dirs,
        libraries=libraries,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        build_ext_class=CUDABuildExtension,
    )


def _clean_build_flags(config_vars: Dict[str, str]) -> None:
    for key, value in config_vars.items():
        if isinstance(value, str):
            value = " " + value + " "
            for s in value.split(" "):
                if (
                    s in ["-Wstrict-prototypes", "-DNDEBUG", "-pg"]
                    or s.startswith("-O")
                    or s.startswith("-g")
                ):
                    value = value.replace(" " + s + " ", " ")
            config_vars[key] = " ".join(value.split())


class CUDABuildExtension(build_ext, object):
    # Refs:
    #   - https://github.com/pytorch/pytorch/torch/utils/cpp_extension.py
    #   - https://github.com/rmcgibbo/npcuda-example/blob/master/cython/setup.py
    #
    def build_extensions(self) -> None:
        # Register .cu  source extensions
        self.compiler.src_extensions.append(".cu")

        # Save references to the original methods
        original_compile = self.compiler._compile

        def cuda_compile(obj, src, ext, cc_args, extra_postargs, pp_opts):
            original_compiler_so = self.compiler.compiler_so
            # Copy before we make any modifications.
            cflags = copy.deepcopy(extra_postargs)
            try:
                if os.path.splitext(src)[-1] == ".cu":
                    if core_defs.CUPY_DEVICE_TYPE == core_defs.DeviceType.ROCM:
                        cuda_exec = os.path.join(gt_config.build_settings["cuda_bin_path"], "hipcc")
                    else:
                        cuda_exec = os.path.join(gt_config.build_settings["cuda_bin_path"], "nvcc")
                    self.compiler.set_executable("compiler_so", [cuda_exec])
                    if isinstance(cflags, dict):
                        cflags = cflags["cuda"]
                elif isinstance(cflags, dict):
                    cflags = cflags["cxx"]

                original_compile(obj, src, ext, cc_args, cflags, pp_opts)
            finally:
                # Put the original compiler back in place.
                self.compiler.set_executable("compiler_so", original_compiler_so)

        self.compiler._compile = cuda_compile
        build_ext.build_extensions(self)
        self.compiler._compile = original_compile



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import collections
import dataclasses
import typing
from typing import Any, ClassVar, Optional, Sequence

import dace
import numpy as np

from gt4py.next import backend as gtx_backend, common as gtx_common
from gt4py.next.ffront import decorator
from gt4py.next.iterator import ir as itir, transforms as itir_transforms
from gt4py.next.iterator.transforms import extractors as extractors
from gt4py.next.otf import arguments, recipes, toolchain
from gt4py.next.program_processors.runners.dace import sdfg_args as gtx_dace_args
from gt4py.next.type_system import type_specifications as ts


@dataclasses.dataclass(frozen=True)
class Program(decorator.Program, dace.frontend.python.common.SDFGConvertible):
    """Extension of GT4Py Program implementing the SDFGConvertible interface via GTIR."""

    sdfg_closure_cache: dict[str, Any] = dataclasses.field(default_factory=dict)
    # Being a ClassVar ensures that in an SDFG with multiple nested GT4Py Programs,
    # there is no name mangling of the connectivity tables used across the nested SDFGs
    # since they share the same memory address.
    connectivity_tables_data_descriptors: ClassVar[
        dict[str, dace.data.Array]
    ] = {}  # symbolically defined

    def __sdfg__(self, *args: Any, **kwargs: Any) -> dace.sdfg.sdfg.SDFG:
        if (self.backend is None) or "dace" not in self.backend.name.lower():
            raise ValueError("The SDFG can be generated only for the DaCe backend.")

        offset_provider: gtx_common.OffsetProvider = self.compilation_options.connectivities or {}
        column_axis = kwargs.get("column_axis", None)

        # TODO(ricoh): connectivity tables required here for now.
        gtir_stage = typing.cast(gtx_backend.Transforms, self.backend.transforms).past_to_itir(
            toolchain.ConcreteArtifact(
                data=self.past_stage,
                args=arguments.CompileTimeArgs(
                    args=tuple(p.type for p in self.past_stage.past_node.params),
                    kwargs={},
                    column_axis=column_axis,
                    offset_provider=offset_provider,
                    argument_descriptor_contexts={},
                ),
            )
        )
        program = gtir_stage.data
        program = itir_transforms.apply_fieldview_transforms(  # run the transforms separately because they require the runtime info
            program, offset_provider=offset_provider
        )
        object.__setattr__(
            gtir_stage,
            "data",
            program,
        )
        object.__setattr__(
            gtir_stage.args, "offset_provider", gtir_stage.args.offset_provider_type
        )  # TODO(ricoh): currently this is circumventing the frozenness of CompileTimeArgs
        # in order to isolate DaCe from the runtime tables in connectivities.offset_provider.
        # These are needed at the time of writing for mandatory GTIR passes.
        # Remove this as soon as Program does not expect connectivity tables anymore.

        _crosscheck_dace_parsing(
            dace_parsed_args=[*args, *kwargs.values()],
            gt4py_program_args=[p.type for p in program.params],
        )

        compile_workflow = typing.cast(
            recipes.OTFCompileWorkflow,
            self.backend.executor
            if not hasattr(self.backend.executor, "step")
            else self.backend.executor.step,
        )  # We know which backend we are using, but we don't know if the compile workflow is cached.
        compile_workflow_translation = (
            compile_workflow.translation
            if not hasattr(compile_workflow.translation, "step")
            else compile_workflow.translation.step
        )  # Same for the translation stage, which could be a `CachedStep` depending on backend configuration.
        # TODO(ricoh): switch 'disable_itir_transforms=True' because we ran them separately previously
        # and so we can ensure the SDFG does not know any runtime info it shouldn't know. Remove with
        # the other parts of the workaround when possible.
        sdfg = dace.SDFG.from_json(
            compile_workflow_translation.replace(
                disable_itir_transforms=True,
                disable_field_origin_on_program_arguments=True,
                use_metrics=False,
            )(gtir_stage).source_code
        )

        self.sdfg_closure_cache["arrays"] = sdfg.arrays

        # Halo exchange related metadata, i.e. gt4py_program_input_fields, gt4py_program_output_fields,
        # offset_providers_per_input_field. Add them as dynamic attributes to the SDFG
        field_params = {
            str(param.id): param for param in program.params if isinstance(param.type, ts.FieldType)
        }

        def single_horizontal_dim_per_field(
            fields: typing.Iterable[itir.Sym],
        ) -> typing.Iterator[tuple[str, gtx_common.Dimension]]:
            for field in fields:
                assert isinstance(field.type, ts.FieldType)
                horizontal_dims = [
                    dim
                    for dim in field.type.dims
                    if dim.kind is gtx_common.DimensionKind.HORIZONTAL
                ]
                # do nothing for fields with multiple horizontal dimensions
                # or without horizontal dimensions
                # this is only meant for use with unstructured grids
                if len(horizontal_dims) == 1:
                    yield str(field.id), horizontal_dims[0]

        input_fields = (
            field_params[name] for name in extractors.InputNamesExtractor.only_fields(program)
        )
        sdfg.gt4py_program_input_fields = dict(single_horizontal_dim_per_field(input_fields))

        output_fields = (
            field_params[name] for name in extractors.OutputNamesExtractor.only_fields(program)
        )
        sdfg.gt4py_program_output_fields = dict(single_horizontal_dim_per_field(output_fields))

        # TODO (ricoh): bring back sdfg.offset_providers_per_input_field.
        #               A starting point would be to use the "trace_shifts" pass on GTIR
        #               and associate the extracted shifts with each input field.
        #               Analogous to the version in `runners.dace_iterator.__init__`, which
        #               was removed when merging #1742.

        return sdfg

    def __sdfg_closure__(self, reevaluate: Optional[dict[str, str]] = None) -> dict[str, Any]:
        """
        Return the closure arrays of the SDFG represented by this object
        as a mapping between array name and the corresponding value.

        The connectivity tables are defined symbolically, i.e. table sizes & strides are DaCe symbols.
        The need to define the connectivity tables in the `__sdfg_closure__` arises from the fact that
        the offset providers are not part of GT4Py Program's arguments.
        Keep in mind, that `__sdfg_closure__` is called after `__sdfg__` method.
        """
        if not self.compilation_options.connectivities:
            return {}

        used_connectivities: dict[str, gtx_common.NeighborConnectivity] = {
            conn_id: conn
            for offset, conn in self.compilation_options.connectivities.items()
            if gtx_common.is_neighbor_table(conn)
            and (conn_id := gtx_dace_args.connectivity_identifier(offset))
            in self.sdfg_closure_cache["arrays"]
        }

        # Define the storage location (e.g. CPU, GPU) of the connectivity tables
        if "storage" not in self.connectivity_tables_data_descriptors:
            for conn_id in used_connectivities.keys():
                self.connectivity_tables_data_descriptors["storage"] = self.sdfg_closure_cache[
                    "arrays"
                ][conn_id].storage
                break

        # Build the closure dictionary
        closure_dict: dict[str, dace.data.Array] = {}
        offset_provider_type = gtx_common.offset_provider_to_type(
            self.compilation_options.connectivities
        )
        for conn_id, conn in used_connectivities.items():
            if conn_id not in self.connectivity_tables_data_descriptors:
                self.connectivity_tables_data_descriptors[conn_id] = dace.data.Array(
                    dtype=dace.dtypes.dtype_to_typeclass(conn.dtype.dtype.type),
                    shape=[
                        gtx_dace_args.field_size_symbol(
                            conn_id, conn.domain.dims[0], offset_provider_type
                        ),
                        gtx_dace_args.field_size_symbol(
                            conn_id, conn.domain.dims[1], offset_provider_type
                        ),
                    ],
                    strides=[
                        gtx_dace_args.field_stride_symbol(
                            conn_id, conn.domain.dims[0], offset_provider_type
                        ),
                        gtx_dace_args.field_stride_symbol(
                            conn_id, conn.domain.dims[1], offset_provider_type
                        ),
                    ],
                    storage=Program.connectivity_tables_data_descriptors["storage"],
                )
            closure_dict[conn_id] = self.connectivity_tables_data_descriptors[conn_id]

        return closure_dict

    def __sdfg_signature__(self) -> tuple[Sequence[str], Sequence[str]]:
        return [p.id for p in self.past_stage.past_node.params], []


def _crosscheck_dace_parsing(dace_parsed_args: list[Any], gt4py_program_args: list[Any]) -> None:
    for dace_parsed_arg, gt4py_program_arg in zip(
        dace_parsed_args,
        gt4py_program_args,
        strict=False,  # dace does not see implicit size args
    ):
        match dace_parsed_arg:
            case dace.data.Scalar():
                assert dace_parsed_arg.dtype == gtx_dace_args.as_dace_type(gt4py_program_arg)
            case bool() | np.bool_():
                assert isinstance(gt4py_program_arg, ts.ScalarType)
                assert gt4py_program_arg.kind == ts.ScalarKind.BOOL
            case int() | np.integer():
                assert isinstance(gt4py_program_arg, ts.ScalarType)
                assert gt4py_program_arg.kind in [ts.ScalarKind.INT32, ts.ScalarKind.INT64]
            case float() | np.floating():
                assert isinstance(gt4py_program_arg, ts.ScalarType)
                assert gt4py_program_arg.kind in [ts.ScalarKind.FLOAT32, ts.ScalarKind.FLOAT64]
            case str() | np.str_():
                assert isinstance(gt4py_program_arg, ts.ScalarType)
                assert gt4py_program_arg.kind == ts.ScalarKind.STRING
            case dace.data.Array():
                assert isinstance(gt4py_program_arg, ts.FieldType)
                assert isinstance(gt4py_program_arg.dtype, ts.ScalarType)
                assert len(dace_parsed_arg.shape) == len(gt4py_program_arg.dims)
                assert dace_parsed_arg.dtype == gtx_dace_args.as_dace_type(gt4py_program_arg.dtype)
            case dace.data.Structure() | dict() | collections.OrderedDict():
                # offset provider
                pass
            case _:
                raise ValueError(
                    f"Unresolved case for {dace_parsed_arg} (==, !=) {gt4py_program_arg}"
                )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import functools
import inspect
import pkgutil
import typing
from collections.abc import Callable

from gt4py.eve import extended_typing as xtyping
from gt4py.eve.extended_typing import (
    Any,
    Final,
    Mapping,
    NestedTuple,
    TypeAlias,
    TypeGuard,
    TypeVar,
)
from gt4py.next import common, utils
from gt4py.next.type_system import type_info, type_specifications as ts


class CustomDataclassNamedCollectionABC(xtyping.DataclassABC):
    """ABC for dataclasses satisfying the current constraints to be custom named collections."""

    @classmethod
    def __subclasshook__(cls, subclass: type) -> bool:
        if dataclasses.is_dataclass(subclass) and (
            not subclass.__module__.startswith("gt4py.")  # Exclude GT4Py internal dataclasses
        ):
            fields = subclass.__dataclass_fields__.values()
            return len(fields) > 0 and all(
                f.init is True
                and f.default is dataclasses.MISSING
                and f.default_factory is dataclasses.MISSING
                and f._field_type is dataclasses._FIELD  # type: ignore[attr-defined]  # private member
                for f in fields
            )

        return False


CustomNamedCollection: TypeAlias = xtyping.TypedNamedTupleABC | CustomDataclassNamedCollectionABC
CUSTOM_NAMED_COLLECTION_TYPES: Final[tuple[type, ...]] = xtyping.get_represented_types(
    CustomNamedCollection
)

NamedCollection: TypeAlias = NestedTuple | CustomNamedCollection
NamedCollectionT = TypeVar("NamedCollectionT", bound=NamedCollection)
NAMED_COLLECTION_TYPES: Final[tuple[type, ...]] = xtyping.get_represented_types(NamedCollection)


NamedCollectionExtractor: TypeAlias = Callable[
    [NamedCollectionT], NestedTuple[common.PrimitiveValue]
]
NamedCollectionConstructor: TypeAlias = Callable[
    [NestedTuple[common.PrimitiveValue]], NamedCollectionT
]

NamedCollectionKey = str | int


def named_collection_type(type_hint: xtyping.TypeAnnotation) -> type[NamedCollection] | None:
    """Get the type if the given type hint represents a supported Python named collection type."""
    class_ = xtyping.get_origin(type_hint) or type_hint
    if class_ is tuple or (
        isinstance(class_, type) and issubclass(class_, CUSTOM_NAMED_COLLECTION_TYPES)
    ):
        return typing.cast(type[NamedCollection], class_)
    return None


def is_named_collection_type(type_hint: xtyping.TypeAnnotation) -> TypeGuard[type[NamedCollection]]:
    """Check if a type annotation represents a supported Python named collection type."""
    return named_collection_type(type_hint) is not None


def elements_keys(
    named_collection_type_hint: xtyping.SingleTypeAnnotation,
) -> tuple[NamedCollectionKey, ...]:
    """Get the keys of the elements in a named collection type."""
    class_ = xtyping.get_origin(named_collection_type_hint) or named_collection_type_hint
    if class_ is tuple:
        return tuple(range(len(xtyping.get_args(named_collection_type_hint))))

    # TODO(egparedes): consider using "__match_args__" as general custom named collection marker
    return tuple(getattr(class_, "__annotations__", {}).keys()) if isinstance(class_, type) else ()


def elements_types(
    named_collection_type_hint: xtyping.SingleTypeAnnotation,
    *,
    globalns: dict[str, Any] | None = None,
    localns: dict[str, Any] | None = None,
) -> Mapping[NamedCollectionKey, type]:
    """Get the types of the elements of a named collection type."""

    if xtyping.get_origin(named_collection_type_hint) is tuple:
        return {i: value for i, value in enumerate(xtyping.get_args(named_collection_type_hint))}

    type_ = named_collection_type(named_collection_type_hint)
    if type_ is not None:
        keys = elements_keys(named_collection_type_hint)
        all_hints = xtyping.get_type_hints(
            named_collection_type_hint, globalns=globalns, localns=localns
        )
        if not {*keys} <= all_hints.keys():
            raise TypeError(
                f"Missing type hints for named collection elements: {keys - all_hints.keys()}"
            )
        assert all(isinstance(k, str) for k in keys)
        return {key: all_hints[key] for key in typing.cast(dict[str, type], keys)}

    return {}


@functools.cache
def make_named_collection_extractor(
    named_collection_type_hint: xtyping.TypeAnnotation,
) -> NamedCollectionExtractor:
    """
    Create an extractor function for the given named collection type.

    The returned function takes an instance of the specified named collection type
    and extracts its values into a nested tuple according to the provided
    type specification.
    """

    extractor_func_src = f"lambda x: {make_extractor_expr(named_collection_type_hint, 'x')}"
    return eval(extractor_func_src)


def make_named_collection_extractor_from_type_spec(
    named_collection_type_spec: ts.NamedCollectionType,
) -> NamedCollectionExtractor:
    """Create an extractor function for the given named collection type specification."""
    assert isinstance(named_collection_type_spec, ts.NamedCollectionType)
    extractor_func_src = (
        f"lambda x: {make_extractor_expr_from_type_spec(named_collection_type_spec, 'x')}"
    )
    return eval(extractor_func_src)


def make_extractor_expr(named_collection_type_hint: xtyping.TypeAnnotation, value_expr: str) -> str:
    """Make an expression from a type description to extract numeric values out of its instances."""

    children_type_hints: dict[str, xtyping.TypeAnnotation] = {}
    expr_parts: list[str] = []
    actual_type = xtyping.get_origin(named_collection_type_hint) or named_collection_type_hint

    if isinstance(actual_type, type):
        if issubclass(actual_type, CUSTOM_NAMED_COLLECTION_TYPES):
            children_type_hints = xtyping.get_type_hints(named_collection_type_hint)
            assert len(children_type_hints)
            expr_parts = [
                make_extractor_expr(value, f"{value_expr}.{key}")
                for key, value in children_type_hints.items()
            ]
        elif issubclass(actual_type, tuple) and (
            tuple_arg_hints := xtyping.get_args(named_collection_type_hint)
        ):
            # This is a `tuple` with type arguments
            expr_parts = [
                make_extractor_expr(value, f"{value_expr}[{i}]")
                for i, value in enumerate(tuple_arg_hints)
            ]

    # Check if we can just skip extraction if the input type
    # is already a pure nested tuple and we just return it as is.
    if isinstance(actual_type, type) and issubclass(actual_type, tuple):
        if (
            children_type_hints
            and all(
                value == f"{value_expr}.{key}"
                for key, value in zip(children_type_hints, expr_parts, strict=True)
            )
        ) or (
            not children_type_hints
            and all(value == f"{value_expr}[{i}]" for i, value in enumerate(expr_parts))
        ):
            return f"{value_expr}"

    # If we don't have information about the inner structure of the type,
    # assume it doesn't need extraction and just return it as is.
    return f"({', '.join(expr_parts)},)" if expr_parts else value_expr


def make_extractor_expr_from_type_spec(type_: ts.TypeSpec, value_expr: str) -> str:
    """Create an expression string that extracts values from a named collection."""
    match type_:
        case ts.NamedCollectionType():
            return make_extractor_expr(pkgutil.resolve_name(type_.original_python_type), value_expr)
        case ts.TupleType() if type_info.needs_value_extraction(type_):
            args = [
                make_extractor_expr_from_type_spec(t, f"{value_expr}[{i}]")
                for i, t in enumerate(type_.types)
            ]
            return f"({', '.join(args)})"

    return value_expr


def _get_named_collection_constructor_args_info(
    named_collection_type_hint: xtyping.SingleTypeAnnotation,
) -> tuple[int, list[str]]:
    if xtyping.get_origin(named_collection_type_hint) is tuple:
        # For plain tuples, we assume all arguments are positional
        args_count = len(xtyping.get_args(named_collection_type_hint))
        return args_count, []

    assert isinstance(named_collection_type_hint, type)
    assert issubclass(named_collection_type_hint, CUSTOM_NAMED_COLLECTION_TYPES)

    # Use a constructor signature without variadic parameters
    for method_name in ("__new__", "__init__"):
        params = inspect.signature(getattr(named_collection_type_hint, method_name)).parameters
        if not (
            {p.kind for p in params.values()}
            & {inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD}
        ):
            break
    else:
        raise TypeError(
            f"Type {named_collection_type_hint} with variadic parameters in constructor are not supported"
        )
    if len(params) == 1:
        raise TypeError(f"Type {named_collection_type_hint} has no parameters in constructor")

    actual_params: dict[str, inspect.Parameter] = {
        k: v for k, v in [*params.items()][1:]
    }  # drop 'self' / 'cls' parameter
    args_count = sum(
        p.kind in {inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD}
        for p in actual_params.values()
    )
    kwargs_keys = [
        key for key, param in actual_params.items() if param.kind == inspect.Parameter.KEYWORD_ONLY
    ]
    assert args_count + len(kwargs_keys) == len(actual_params), (
        "Variadic parameters are not supported"
    )

    return args_count, kwargs_keys


@functools.cache
def make_named_collection_constructor(
    named_collection_type_hint: type[NamedCollectionT] | xtyping.TypeAnnotation,
    *,
    nested: bool = True,
) -> NamedCollectionConstructor[NamedCollectionT]:
    """
    Create a constructor function for the given named collection type.

    The returned function takes a nested tuple of values and constructs
    an instance of the specified named collection type.
    """
    assert (type_ := named_collection_type(named_collection_type_hint)) is not None
    assert issubclass(type_, NAMED_COLLECTION_TYPES)

    global_ns: dict[str, Any] = {}
    constructor_func_body = make_constructor_expr(
        named_collection_type_hint, "x", item_path="", global_ns=global_ns, nested=nested
    )
    constructor_func_src = f"lambda x: {constructor_func_body}"

    return eval(constructor_func_src, global_ns)


def make_named_collection_constructor_from_type_spec(
    named_collection_type_spec: ts.NamedCollectionType, nested: bool = True
) -> NamedCollectionConstructor:
    """Create a constructor function for the given named collection type specification."""
    assert isinstance(named_collection_type_spec, ts.NamedCollectionType)
    return make_named_collection_constructor(
        pkgutil.resolve_name(named_collection_type_spec.original_python_type), nested=nested
    )


def make_constructor_expr(
    named_collection_type_hint: xtyping.TypeAnnotation,
    value_expr: str,
    *,
    item_path: str,
    global_ns: dict[str, Any],
    nested: bool = True,  # TODO: consider splitting into separate function
) -> str:
    """
    Create an expression string that constructs a named collection from a nested tuple of values.

    The expression assumes that `value_expr` is a nested tuple of values that
    correspond to the structure of `named collection_type`. The expression constructs
    an instance of `named collection_type` using these values.

    Arguments:
        named_collection_type_hint: The type of the named collection to construct.
        value_expr: The expression that evaluates to the nested tuple of values.

    Keyword Arguments:
        item_path: A unique path string used to create unique names for named collection
            types in the constructor building expression.
        global_ns: The global namespace where the unique named collection type aliases are
            stored for final evaluation of the constructor expression.
    """
    actual_type = xtyping.get_origin(named_collection_type_hint) or named_collection_type_hint
    nested_types: dict[int | str, xtyping.TypeAnnotation] = {}
    if isinstance(actual_type, type):
        if issubclass(actual_type, CUSTOM_NAMED_COLLECTION_TYPES):
            # Store the named collection type alias in the global namespace for eval()
            named_collection_type_alias = f"__{item_path}__{actual_type.__name__}"
            global_ns[named_collection_type_alias] = actual_type

            # Get the type hints of the named collection's members
            named_collection_type = typing.cast(str | type, named_collection_type_hint)
            named_collection_keys = elements_keys(named_collection_type)
            type_hints = xtyping.get_type_hints(named_collection_type)
            assert {*named_collection_keys} <= type_hints.keys(), (
                "Mismatch between keys and type hints"
            )
            nested_types = {
                key: type_hints[key] for key in typing.cast(dict[str, type], named_collection_keys)
            }

        elif issubclass(actual_type, tuple) and (
            tuple_args_hint := xtyping.get_args(named_collection_type_hint)
        ):
            named_collection_type_alias = ""
            nested_types = {i: type_hint for i, type_hint in enumerate(tuple_args_hint)}

    if nested_types:
        if nested:
            call_args = [
                make_constructor_expr(
                    type_hint,
                    f"{value_expr}[{i}]",
                    item_path=f"{item_path}__{key}",
                    global_ns=global_ns,
                )
                for i, (key, type_hint) in enumerate(nested_types.items())
            ]
        else:
            call_args = [f"{value_expr}[{i}]" for i in range(len(nested_types))]

        # Optimize the call expression if none of the children needs further construction
        # use argument unpacking to pass them to the constructor
        if all(arg == f"{value_expr}[{i}]" for i, arg in enumerate(call_args)):
            if not named_collection_type_alias:
                # Fast path for plain tuples: none of the children needs construction, so we can just
                # return the value expression as is
                return value_expr

            # Fast path: none of the children needs construction, so we can just use argument
            # unpacking to pass them to the constructor
            args_count, kwargs_keys = _get_named_collection_constructor_args_info(
                named_collection_type_hint
            )
            if not kwargs_keys:
                call_args = [f"*{value_expr}"]
            else:
                call_args = [
                    f"*{value_expr}[:{args_count}]",
                    *(f"{k}={value_expr}[{args_count + i}]" for i, k in enumerate(kwargs_keys)),
                ]

        return f"{named_collection_type_alias}({', '.join(call_args)})"

    # We don't need to construct anything, just return the argument as is
    return value_expr


# typing of this function is sketchy
def tree_map_named_collection(
    fun: Callable[utils._P, utils._R],
) -> Callable[..., utils._R | NamedCollection]:  # nested NamedCollection[_R]
    # TODO(havogt): if we figure out the concept of the collection category that tree_map accepts, we could simplify this implementation,
    # currently `tree_map` requires `Iterable` collections, therefore we have to manually extract/construct here.
    def impl(
        *args: Any | NamedCollection,
    ) -> utils._R | NamedCollection:  # type annotation for human...
        # TODO assert all inputs have the same (nested) collection structure
        if is_named_collection_type(type(args[0])):
            extract = make_named_collection_extractor(type(args[0]))  # type: ignore[arg-type]
            construct = make_named_collection_constructor(type(args[0]), nested=True)  # type: ignore[arg-type]
        else:
            extract = construct = lambda x: x
        plain_tuples = (extract(arg) for arg in args)
        result = utils.tree_map(fun=fun)(*plain_tuples)
        return construct(result)  # type: ignore[arg-type]

    return impl



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from . import attrib, meta
from .base import (
    NOTHING,
    BaseFrozen,
    BaseSingleton,
    Registry,
    UniqueIdGenerator,
    classmethod_to_function,
    classproperty,
    compose,
    flatten,
    flatten_iter,
    get_member,
    is_collection,
    is_identifier_name,
    is_iterable_of,
    is_mapping_of,
    jsonify,
    listify,
    make_dir,
    make_local_dir,
    make_module_from_file,
    namespace_from_nested_dict,
    normalize_mapping,
    patch_module,
    restore_module,
    shash,
    shashed_id,
    slugify,
    warn_experimental_feature,
)
from .field import Field


__all__ = [  # noqa: RUF022 `__all__` is not sorted
    # Modules
    "attrib",
    "meta",
    # Objects
    "NOTHING",
    "BaseFrozen",
    "BaseSingleton",
    "Registry",
    "UniqueIdGenerator",
    "classmethod_to_function",
    "classproperty",
    "compose",
    "Field",
    "flatten",
    "flatten_iter",
    "get_member",
    "is_collection",
    "is_identifier_name",
    "is_iterable_of",
    "is_mapping_of",
    "jsonify",
    "listify",
    "make_dir",
    "make_local_dir",
    "make_module_from_file",
    "namespace_from_nested_dict",
    "normalize_mapping",
    "patch_module",
    "restore_module",
    "shash",
    "shashed_id",
    "slugify",
    "warn_experimental_feature",
]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause
from __future__ import annotations

import os
import pathlib
import pickle
from typing import Any, Hashable

from gt4py._core import locking
from gt4py.eve import utils as eve_utils


class FileCache:
    """
    Dictionary-like object for persistently caching objects on disk.
    """

    path: pathlib.Path

    def __init__(self, path: str | os.PathLike):
        self.path = pathlib.Path(path).resolve()
        self.path.mkdir(parents=True, exist_ok=True)

    def _get_path(self, key: Hashable) -> pathlib.Path:
        """Return the path where an item with `key` is stored."""
        key = eve_utils.content_hash(key)
        path = self.path / f"{key}.pkl"
        return path

    def __getitem__(self, key: Hashable) -> Any:
        if key not in self:
            raise KeyError(key)
        with locking.lock(path := self._get_path(key)):
            with open(path, "rb") as f:
                return pickle.load(f)

    def __setitem__(self, key: Hashable, value: Any) -> None:
        with locking.lock(path := self._get_path(key)):
            with open(path, "wb") as f:
                pickle.dump(value, f, protocol=5)

    def __delitem__(self, key: Hashable) -> None:
        if key not in self:
            raise KeyError(key)
        with locking.lock(path := self._get_path(key)):
            path.unlink()

    def __contains__(self, key: Hashable) -> bool:
        return self._get_path(key).exists()



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from typing import Protocol, TypeAlias, TypeVar

from gt4py.next.ffront import stages as ffront_stages
from gt4py.next.iterator import ir as itir
from gt4py.next.otf import arguments, code_specs, stages, toolchain, workflow


CodeSpecT = TypeVar("CodeSpecT", bound=code_specs.SourceCodeSpec)
TargetCodeSpecT = TypeVar("TargetCodeSpecT", bound=code_specs.SourceCodeSpec)


IRDefinitionT = TypeVar(
    "IRDefinitionT",
    ffront_stages.DSLFieldOperatorDef,
    ffront_stages.DSLProgramDef,
    ffront_stages.FOASTOperatorDef,
    ffront_stages.PASTProgramDef,
    itir.Program,
)
ArgsDefinitionT = TypeVar("ArgsDefinitionT", arguments.JITArgs, arguments.CompileTimeArgs)

ConcreteProgramDef: TypeAlias = toolchain.ConcreteArtifact[IRDefinitionT, ArgsDefinitionT]
CompilableProgramDef: TypeAlias = ConcreteProgramDef[itir.Program, arguments.CompileTimeArgs]


class TranslationStep(
    workflow.ReplaceEnabledWorkflowMixin[CompilableProgramDef, stages.ProgramSource[CodeSpecT]],
    Protocol[CodeSpecT],
):
    """Translate a GT4Py program to source code (ProgramCall -> ProgramSource)."""

    ...


class BindingStep(Protocol[CodeSpecT, TargetCodeSpecT]):
    """
    Generate Bindings for program source and package both together (ProgramSource -> CompilableSource).

    In the special cases where bindings are not required, such a step could also simply construct
    a ``CompilableSource`` from the ``ProgramSource`` with bindings set to ``None``.
    """

    def __call__(
        self, program_source: stages.ProgramSource[CodeSpecT]
    ) -> stages.CompilableProject[CodeSpecT, TargetCodeSpecT]: ...


class CompilationStep(
    workflow.Workflow[
        stages.CompilableProject[CodeSpecT, TargetCodeSpecT], stages.ExecutableProgram
    ],
    Protocol[CodeSpecT, TargetCodeSpecT],
):
    """Compile program source code and bindings into a python callable (CompilableSource -> CompiledProgram)."""

    def __call__(
        self, source: stages.CompilableProject[CodeSpecT, TargetCodeSpecT]
    ) -> stages.ExecutableProgram: ...



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import functools
import importlib.util
import pathlib
import tempfile
import textwrap
import typing
from collections.abc import Callable, Iterable
from typing import Any, Optional

from gt4py.eve import codegen
from gt4py.eve.codegen import FormatTemplate as as_fmt, MakoTemplate as as_mako
from gt4py.next import (
    backend as next_backend,
    common,
    config,
    custom_layout_allocators as next_allocators,
)
from gt4py.next.ffront import foast_to_gtir, foast_to_past, past_to_itir
from gt4py.next.iterator import ir as itir, transforms as itir_transforms
from gt4py.next.otf import definitions, stages, workflow
from gt4py.next.type_system import type_info, type_specifications as ts


def _create_tmp(axes: str, origin: str, shape: str, dtype: ts.TypeSpec) -> str:
    if isinstance(dtype, ts.TupleType):
        return f"({','.join(_create_tmp(axes, origin, shape, dt) for dt in dtype.types)},)"
    else:
        assert isinstance(dtype, ts.ScalarType)
        return (
            f"gtx.as_field([{axes}], np.empty({shape}, dtype=np.dtype('{dtype}')), origin={origin})"
        )


class EmbeddedDSL(codegen.TemplatedGenerator):
    Sym = as_fmt("{id}")
    SymRef = as_fmt("{id}")

    def visit_Literal(self, node: itir.Literal, **kwargs: Any) -> str:
        if (
            isinstance(node.type, ts.ScalarType)
            and type_info.is_floating_point(node.type)
            and node.value in ["inf", "-inf", "nan"]
        ):
            dtype = node.type.kind.name.lower()
            if node.value == "inf":
                return f"np.{dtype}(np.inf)"
            elif node.value == "-inf":
                return f"-np.{dtype}(np.inf)"
            elif node.value == "nan":
                return f"np.{dtype}(np.nan)"
        return node.value

    NoneLiteral = as_fmt("None")
    OffsetLiteral = as_fmt("{value}")
    AxisLiteral = as_fmt("{value}")
    FunCall = as_fmt("{fun}({','.join(args)})")
    Lambda = as_mako("(lambda ${','.join(params)}: ${expr})")
    FunctionDefinition = as_mako(
        """
@fundef
def ${id}(${','.join(params)}):
    return ${expr}
    """
    )
    Program = as_mako(
        """
${''.join(function_definitions)}
@fendef
def ${id}(${','.join(params)}):
    ${'\\n    '.join(declarations)}
    ${'\\n    '.join(body)}
    """
    )
    SetAt = as_mako("set_at(${expr}, ${domain}, ${target})")
    IfStmt = as_mako("""if_stmt(${cond}, 
        lambda: [${','.join(true_branch)}],
        lambda: [${','.join(false_branch)}]
    )""")

    def visit_Temporary(self, node: itir.Temporary, **kwargs: Any) -> str:
        assert (
            isinstance(node.domain, itir.FunCall)
            and isinstance(node.domain.fun, itir.SymRef)
            and node.domain.fun.id in ("cartesian_domain", "unstructured_domain")
        )
        assert all(
            isinstance(r, itir.FunCall) and r.fun == itir.SymRef(id="named_range")
            for r in node.domain.args
        )
        domain_ranges = [self.visit(r.args) for r in node.domain.args]  # type: ignore[attr-defined] # `node.domain` is `FunCall` checked in previous assert
        axes = ", ".join(label for label, _, _ in domain_ranges)
        origin = "{" + ", ".join(f"{label}: -{start}" for label, start, _ in domain_ranges) + "}"
        shape = "(" + ", ".join(f"{stop}-{start}" for _, start, stop in domain_ranges) + ")"
        assert node.dtype
        return f"{node.id} = {_create_tmp(axes, origin, shape, node.dtype)}"


_FENCIL_CACHE: dict[int, Callable] = {}


def fencil_generator(
    ir: itir.Program,
    debug: bool,
    use_embedded: bool,
    offset_provider: common.OffsetProvider,
    transforms: itir_transforms.GTIRTransform,
) -> stages.ExecutableProgram:
    """
    Generate a directly executable fencil from an ITIR node.

    Arguments:
        ir: The iterator IR (ITIR) node.
        debug: Keep module source containing fencil implementation.
        extract_temporaries: Extract intermediate field values into temporaries.
        use_embedded: Directly use builtins from embedded backend instead of
                      generic dispatcher. Gives faster performance and is easier
                      to debug.
        offset_provider: A mapping from offset names to offset providers.
    """
    # TODO(tehrengruber): just a temporary solution until we have a proper generic
    #  caching mechanism
    cache_key = hash(
        (
            ir,
            transforms,
            debug,
            use_embedded,
            tuple(common.offset_provider_to_type(offset_provider).items()),
        )
    )
    if cache_key in _FENCIL_CACHE:
        if debug:
            print(f"Using cached fencil for key {cache_key}")
        return _FENCIL_CACHE[cache_key]  # A CompiledProgram is just a Callable

    ir = transforms(ir, offset_provider=offset_provider)

    program = EmbeddedDSL.apply(ir)

    # format output in debug mode for better debuggability
    # (e.g. line numbers, overview in the debugger).
    if debug:
        program = codegen.format_python_source(program)

    offset_literals: Iterable[str] = (
        ir.pre_walk_values()
        .if_isinstance(itir.OffsetLiteral)
        .getattr("value")
        .if_isinstance(str)
        .to_set()
    )
    axis_literals_set: Iterable[itir.AxisLiteral] = (
        ir.pre_walk_values().if_isinstance(itir.AxisLiteral).to_set()
    )

    if use_embedded:
        builtins_import = "from gt4py.next.iterator.embedded import *"
    else:
        builtins_import = "from gt4py.next.iterator.builtins import *"

    header = textwrap.dedent(
        f"""
        import numpy as np
        import gt4py.next as gtx
        {builtins_import}
        from gt4py.next.iterator.runtime import *
        """
    )

    with tempfile.NamedTemporaryFile(
        mode="w", suffix=".py", encoding="utf-8", delete=False
    ) as source_file:
        source_file_name = source_file.name
        if debug:
            print(source_file_name)
        offset_literals = [f'{o} = offset("{o}")' for o in offset_literals]
        axis_literals = [
            f'{o.value} = gtx.Dimension("{o.value}", kind=gtx.DimensionKind("{o.kind}"))'
            for o in axis_literals_set
        ]
        source_file.write(header)
        source_file.write("\n".join(offset_literals))
        source_file.write("\n")
        source_file.write("\n".join(axis_literals))
        source_file.write("\n")
        source_file.write(program)
    try:
        spec = importlib.util.spec_from_file_location("module.name", source_file_name)
        mod = importlib.util.module_from_spec(spec)  # type: ignore
        spec.loader.exec_module(mod)  # type: ignore
    finally:
        if not debug:
            pathlib.Path(source_file_name).unlink(missing_ok=True)

    assert isinstance(ir, itir.Program)
    fencil_name = ir.id
    fencil = getattr(mod, fencil_name)

    _FENCIL_CACHE[cache_key] = fencil

    return typing.cast(stages.ExecutableProgram, fencil)


@dataclasses.dataclass(frozen=True)
class Roundtrip(workflow.Workflow[definitions.CompilableProgramDef, stages.ExecutableProgram]):
    debug: Optional[bool] = None
    use_embedded: bool = True
    dispatch_backend: Optional[next_backend.Backend] = None
    transforms: itir_transforms.GTIRTransform = itir_transforms.apply_common_transforms  # type: ignore[assignment] # TODO(havogt): cleanup interface of `apply_common_transforms`

    def __call__(self, inp: definitions.CompilableProgramDef) -> stages.ExecutableProgram:
        debug = config.DEBUG if self.debug is None else self.debug

        fencil = fencil_generator(
            inp.data,
            offset_provider=inp.args.offset_provider,
            debug=debug,
            use_embedded=self.use_embedded,
            transforms=self.transforms,
        )

        def decorated_fencil(
            *args: Any,
            offset_provider: dict[str, common.Connectivity | common.Dimension],
            out: Any = None,
            column_axis: Optional[common.Dimension] = None,
            **kwargs: Any,
        ) -> None:
            if out is not None:
                args = (*args, out)
            if not column_axis:  # TODO(tehrengruber): This variable is never used. Bug?
                column_axis = inp.args.column_axis
            fencil(
                *args,
                offset_provider=offset_provider,
                backend=self.dispatch_backend,
                column_axis=inp.args.column_axis,
                **kwargs,
            )

        return decorated_fencil


# TODO(tehrengruber): introduce factory
default = next_backend.Backend(
    name="roundtrip",
    executor=Roundtrip(
        transforms=functools.partial(
            itir_transforms.apply_common_transforms,
            extract_temporaries=False,
        )
    ),
    allocator=next_allocators.StandardCPUFieldBufferAllocator(),
    transforms=next_backend.DEFAULT_TRANSFORMS,
)
with_temporaries = next_backend.Backend(
    name="roundtrip_with_temporaries",
    executor=Roundtrip(
        transforms=functools.partial(
            itir_transforms.apply_common_transforms,
            extract_temporaries=True,
        )
    ),
    allocator=next_allocators.StandardCPUFieldBufferAllocator(),
    transforms=next_backend.DEFAULT_TRANSFORMS,
)
no_transforms = next_backend.Backend(
    name="roundtrip",
    executor=Roundtrip(transforms=lambda o, *, offset_provider: o),
    allocator=next_allocators.StandardCPUFieldBufferAllocator(),
    transforms=next_backend.DEFAULT_TRANSFORMS,
)


gtir = next_backend.Backend(
    name="roundtrip_gtir",
    executor=Roundtrip(transforms=itir_transforms.apply_fieldview_transforms),  # type: ignore[arg-type] # don't understand why mypy complains
    allocator=next_allocators.StandardCPUFieldBufferAllocator(),
    transforms=next_backend.Transforms(
        past_to_itir=past_to_itir.past_to_gtir_factory(),
        foast_to_itir=foast_to_gtir.adapted_foast_to_gtir_factory(cached=True),
        field_view_op_to_prog=foast_to_past.operator_to_program_factory(
            foast_to_itir_step=foast_to_gtir.adapted_foast_to_gtir_factory()
        ),
    ),
)
foast_to_gtir_step = foast_to_gtir.adapted_foast_to_gtir_factory(cached=True)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from gt4py import eve
from gt4py.next.iterator import ir as itir
from gt4py.next.type_system import type_specifications as ts


class SymbolNameSetExtractor(eve.NodeVisitor):
    """Extract a set of symbol names"""

    def visit_Literal(self, node: itir.Literal) -> set[str]:
        return set()

    def generic_visitor(self, node: itir.Node) -> set[str]:
        input_fields: set[str] = set()
        for child in eve.trees.iter_children_values(node):
            input_fields |= self.visit(child)
        return input_fields

    def visit_Node(self, node: itir.Node) -> set[str]:
        return set()

    def visit_Program(self, node: itir.Program) -> set[str]:
        names = set()
        for stmt in node.body:
            names |= self.visit(stmt)
        return names

    def visit_IfStmt(self, node: itir.IfStmt) -> set[str]:
        names = set()
        for stmt in node.true_branch + node.false_branch:
            names |= self.visit(stmt)
        return names

    def visit_Temporary(self, node: itir.Temporary) -> set[str]:
        return set()

    def visit_SymRef(self, node: itir.SymRef) -> set[str]:
        return {str(node.id)}

    @classmethod
    def only_fields(cls, program: itir.Program) -> set[str]:
        field_param_names = [
            str(param.id) for param in program.params if isinstance(param.type, ts.FieldType)
        ]
        return {name for name in cls().visit(program) if name in field_param_names}


class InputNamesExtractor(SymbolNameSetExtractor):
    """Extract the set of symbol names passed into field operators within a program."""

    def visit_SetAt(self, node: itir.SetAt) -> set[str]:
        return self.visit(node.expr)

    def visit_FunCall(self, node: itir.FunCall) -> set[str]:
        input_fields = set()
        for arg in node.args:
            input_fields |= self.visit(arg)
        return input_fields


class OutputNamesExtractor(SymbolNameSetExtractor):
    """Extract the set of symbol names written to within a program"""

    def visit_SetAt(self, node: itir.SetAt) -> set[str]:
        return self.visit(node.target)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""
Typing definitions working across different Python versions (via `typing_extensions`).

Definitions in 'typing_extensions' take priority over those in 'typing'.
"""

from __future__ import annotations

# ruff: noqa: F401, F405
import abc as _abc
import array as _array
import collections.abc as _collections_abc
import dataclasses as _dataclasses
import functools as _functools
import inspect as _inspect
import mmap as _mmap
import pickle as _pickle
import sys as _sys
import types as _types
import typing as _typing
from typing import *  # noqa: F403 [undefined-local-with-import-star]
from typing import overload

import numpy.typing as npt
import typing_extensions as _typing_extensions
from typing_extensions import *  # type: ignore[assignment,no-redef]  # noqa: F403 [undefined-local-with-import-star]


if _sys.version_info >= (3, 9):
    # Standard library already supports PEP 585 (Type Hinting Generics In Standard Collections)
    from builtins import (  # type: ignore[assignment]
        dict as Dict,
        frozenset as FrozenSet,
        list as List,
        set as Set,
        tuple as Tuple,
        type as Type,
    )
    from collections import (
        ChainMap as ChainMap,
        Counter as Counter,
        OrderedDict as OrderedDict,
        defaultdict as defaultdict,
        deque as deque,
    )
    from collections.abc import (
        AsyncGenerator as AsyncGenerator,
        AsyncIterable as AsyncIterable,
        AsyncIterator as AsyncIterator,
        Awaitable as Awaitable,
        ByteString as ByteString,
        Callable as Callable,
        Collection as Collection,
        Container as Container,
        Coroutine as Coroutine,
        Generator as Generator,
        ItemsView as ItemsView,
        Iterable as Iterable,
        Iterator as Iterator,
        KeysView as KeysView,
        Mapping as Mapping,
        MappingView as MappingView,
        MutableMapping as MutableMapping,
        MutableSequence as MutableSequence,
        MutableSet as MutableSet,
        Reversible as Reversible,
        Sequence as Sequence,
        Set as AbstractSet,
        ValuesView as ValuesView,
    )
    from contextlib import (
        AbstractAsyncContextManager as AsyncContextManager,
        AbstractContextManager as ContextManager,
    )
    from re import Match as Match, Pattern as Pattern


# These fallbacks are useful for public symbols not exported by default.
# Again, definitions in 'typing_extensions' take priority over those in 'typing'
def __getattr__(name: str) -> Any:
    import sys

    import typing_extensions

    result = SENTINEL = object()
    if not (name.startswith("__") and name.endswith("__")):
        result = getattr(typing_extensions, name, SENTINEL)
        if result is SENTINEL:
            import typing

            result = getattr(typing, name, SENTINEL)

    if result is SENTINEL:
        raise AttributeError(f"module '{__name__}' has no attribute '{name}'")

    setattr(sys.modules[__name__], name, result)  # cache result

    return result


def __dir__() -> List[str]:
    if not hasattr(self_func := (globals()["__dir__"]), "__cached_dir"):
        import typing

        import typing_extensions

        orig_dir = typing.__dir__()
        self_func.__cached_dir = [*orig_dir] + [
            name for name in typing_extensions.__dir__() if name not in orig_dir
        ]

    return self_func.__cached_dir


# -- Common type aliases --
NoArgsCallable = Callable[[], Any]

_A = TypeVar("_A", contravariant=True)
_R = TypeVar("_R", covariant=True)


class ArgsOnlyCallable(Protocol[_A, _R]):
    def __call__(self, *args: _A) -> _R: ...


_T_co = TypeVar("_T_co", covariant=True)
NestedSequence = Sequence[Union[_T_co, "NestedSequence[_T_co]"]]
NestedList = List[Union[_T_co, "NestedList[_T_co]"]]
NestedTuple = Tuple[Union[_T_co, "NestedTuple[_T_co]"], ...]

MaybeNested = Union[_T_co, NestedSequence[_T_co]]
MaybeNestedInSequence = Union[_T_co, NestedSequence[_T_co]]
MaybeNestedInList = Union[_T_co, NestedList[_T_co]]
MaybeNestedInTuple = Union[_T_co, NestedTuple[_T_co]]


def is_nested_tuple_of(value: object, type_: type[_T_co]) -> TypeIs[NestedTuple[_T_co]]:
    """Check if `value` is a nested tuple of elements of type `type_`."""
    return isinstance(value, tuple) and all(
        isinstance(v, type_) or (isinstance(v, tuple) and is_nested_tuple_of(v, type_))
        for v in value
    )


def is_maybe_nested_in_tuple_of(
    value: object, type_: type[_T_co]
) -> TypeIs[MaybeNestedInTuple[_T_co]]:
    """Check if `value` is either of type `type_` or a nested tuple of elements of type `type_`."""
    return isinstance(value, type_) or is_nested_tuple_of(value, type_)


# -- Typing annotations --
SingleTypeAnnotation = Union[
    Type,
    _types.GenericAlias,
    _typing._BaseGenericAlias,  # type: ignore[name-defined]  # _BaseGenericAlias is not exported in stub
]

SolvedTypeAnnotation = Union[SingleTypeAnnotation, _typing._SpecialForm]

TypeAnnotation = Union[ForwardRef, SolvedTypeAnnotation]
SourceTypeAnnotation = Union[str, TypeAnnotation]

StdGenericAliasType: Final[Type] = type(List[int])

if TYPE_CHECKING:
    StdGenericAlias: TypeAlias = _types.GenericAlias

_TypingSpecialFormType: Final[Type] = _typing._SpecialForm
_TypingGenericAliasType: Final[Type] = _typing._BaseGenericAlias  # type: ignore[attr-defined]  # _BaseGenericAlias / _GenericAlias are not exported in stub


# -- Standard Python protocols --
_C = TypeVar("_C")
_V = TypeVar("_V")


class NonDataDescriptor(Protocol[_C, _V]):
    """Typing protocol for non-data descriptor classes.

    See https://docs.python.org/3/howto/descriptor.html for further information.
    """

    @overload
    def __get__(
        self, _instance: Literal[None], _owner_type: Optional[Type[_C]] = None
    ) -> NonDataDescriptor[_C, _V]: ...

    @overload
    def __get__(self, _instance: _C, _owner_type: Optional[Type[_C]] = None) -> _V: ...

    def __get__(
        self, _instance: Optional[_C], _owner_type: Optional[Type[_C]] = None
    ) -> _V | NonDataDescriptor[_C, _V]: ...


class DataDescriptor(NonDataDescriptor[_C, _V], Protocol):
    """Typing protocol for data descriptor classes.

    See https://docs.python.org/3/howto/descriptor.html for further information.
    """

    def __set__(self, _instance: _C, _value: _V) -> None: ...

    def __delete__(self, _instance: _C) -> None: ...


# -- Based on typeshed definitions --
ReadOnlyBuffer: TypeAlias = Union[bytes, SupportsBytes]
WriteableBuffer: TypeAlias = Union[
    bytearray, memoryview, _array.array, _mmap.mmap, _pickle.PickleBuffer
]
ReadableBuffer: TypeAlias = Union[ReadOnlyBuffer, WriteableBuffer]


_P = ParamSpec("_P")
_T = TypeVar("_T")


class SingleDispatchCallable(Protocol[_P, _T]):
    registry: Mapping[Any, Callable[_P, _T]]

    def dispatch(self, cls: Any) -> Callable[_P, _T]: ...

    @overload
    def register(
        self, cls: Any, func: Literal[None] = None
    ) -> Callable[[Callable[_P, _T]], Callable[_P, _T]]: ...

    @overload
    def register(self, cls: Any, func: Callable[_P, _T]) -> Callable[_P, _T]: ...

    def register(
        self, cls: Any, func: Callable[_P, _T] | None = None
    ) -> Callable[[Callable[_P, _T]], Callable[_P, _T]] | Callable[_P, _T]: ...

    def _clear_cache(self) -> None: ...

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _T: ...


def is_single_dispatch_callable(
    func: Callable[_P, _T],
) -> TypeGuard[SingleDispatchCallable[_P, _T]]:
    return (
        callable(func)
        and callable(getattr(func, "dispatch", None))
        and callable(getattr(func, "register", None))
    )


class HashlibAlgorithm(Protocol):
    """Used in the hashlib module of the standard library."""

    @property
    def block_size(self) -> int: ...

    @property
    def digest_size(self) -> int: ...

    @property
    def name(self) -> str: ...

    def __init__(self, data: ReadableBuffer = ...) -> None: ...

    def copy(self) -> Self: ...

    def update(self, data: Buffer, /) -> None: ...

    def digest(self) -> bytes: ...

    def hexdigest(self) -> str: ...


# -- Third party protocols --
class SupportsArray(Protocol):
    def __array__(self, dtype: Optional[npt.DTypeLike] = None, /) -> npt.NDArray[Any]: ...


def supports_array(value: Any) -> TypeGuard[SupportsArray]:
    return hasattr(value, "__array__")


class ArrayInterface(Protocol):
    @property
    def __array_interface__(self) -> Dict[str, Any]: ...


class ArrayInterfaceTypedDict(TypedDict):
    shape: Tuple[int, ...]
    typestr: str
    descr: NotRequired[List[Tuple]]
    data: NotRequired[Tuple[int, bool]]
    strides: NotRequired[Optional[Tuple[int, ...]]]
    mask: NotRequired[Optional["StrictArrayInterface"]]
    offset: NotRequired[int]
    version: int


class StrictArrayInterface(Protocol):
    @property
    def __array_interface__(self) -> ArrayInterfaceTypedDict: ...


def supports_array_interface(value: Any) -> TypeGuard[ArrayInterface]:
    return hasattr(value, "__array_interface__")


class CUDAArrayInterface(Protocol):
    @property
    def __cuda_array_interface__(self) -> Dict[str, Any]: ...


class CUDAArrayInterfaceTypedDict(TypedDict):
    shape: Tuple[int, ...]
    typestr: str
    data: Tuple[int, bool]
    version: int
    strides: NotRequired[Optional[Tuple[int, ...]]]
    descr: NotRequired[List[Tuple]]
    mask: NotRequired[Optional["StrictCUDAArrayInterface"]]
    stream: NotRequired[Optional[int]]


class StrictCUDAArrayInterface(Protocol):
    @property
    def __cuda_array_interface__(self) -> CUDAArrayInterfaceTypedDict: ...


def supports_cuda_array_interface(value: Any) -> TypeGuard[CUDAArrayInterface]:
    """Check if the given value supports the CUDA Array Interface."""
    return hasattr(value, "__cuda_array_interface__")


DLPackDevice = Tuple[int, int]


class MultiStreamDLPackBuffer(Protocol):
    def __dlpack__(self, *, stream: Optional[int] = None) -> Any: ...

    def __dlpack_device__(self) -> DLPackDevice: ...


class SingleStreamDLPackBuffer(Protocol):
    def __dlpack__(self, *, stream: None = None) -> Any: ...

    def __dlpack_device__(self) -> DLPackDevice: ...


DLPackBuffer: TypeAlias = Union[MultiStreamDLPackBuffer, SingleStreamDLPackBuffer]


def supports_dlpack(value: Any) -> TypeGuard[DLPackBuffer]:
    """Check if a given object supports the DLPack protocol."""
    return callable(getattr(value, "__dlpack__", None)) and callable(
        getattr(value, "__dlpack_device__", None)
    )


class DevToolsPrettyPrintable(Protocol):
    """Used by python-devtools (https://python-devtools.helpmanual.io/)."""

    def __pretty__(
        self, fmt: Callable[[Any], Any], **kwargs: Any
    ) -> Generator[Any, None, None]: ...


# -- Added functionality --
_ArtefactTypes: tuple[type, ...] = (_types.GenericAlias,)

# `Any` is a class since Python 3.11
if isinstance(_typing.Any, type):  # Python >= 3.11
    _ArtefactTypes = (*_ArtefactTypes, _typing.Any)

# `Any` is a class since typing_extensions >= 4.4 and Python 3.11
if (typing_exts_any := getattr(_typing_extensions, "Any", None)) is not _typing.Any and isinstance(
    typing_exts_any, type
):
    _ArtefactTypes = (*_ArtefactTypes, typing_exts_any)


def is_actual_type(obj: Any) -> TypeGuard[Type]:
    """Check if an object has an actual type and instead of a typing artefact like ``GenericAlias`` or ``Any``.

    This is needed because since Python 3.9:
        ``isinstance(types.GenericAlias(), type) is True``
    and since Python 3.11:
        ``isinstance(typing.Any, type) is True``
    """
    return (
        isinstance(obj, type) and (obj not in _ArtefactTypes) and (type(obj) not in _ArtefactTypes)
    )


if hasattr(_typing_extensions, "Any") and _typing.Any is not _typing_extensions.Any:  # type: ignore[attr-defined] # _typing_extensions.Any only from >= 4.4
    # When using Python < 3.11 and typing_extensions >= 4.4 there are
    # two different implementations of `Any`

    def is_Any(obj: Any) -> bool:
        return obj is _typing.Any or obj is _typing_extensions.Any  # type: ignore[attr-defined] # _typing_extensions.Any only from >= 4.4

else:

    def is_Any(obj: Any) -> bool:
        return obj is _typing.Any


def has_type_parameters(cls: Type) -> bool:
    """Return ``True`` if obj is a generic class with type parameters."""
    return issubclass(cls, Generic) and len(getattr(cls, "__parameters__", [])) > 0  # type: ignore[arg-type]  # Generic not considered as a class


def get_actual_type(obj: _T) -> Type[_T]:
    """Return type of an object (also working for GenericAlias instances which pretend to be an actual type)."""
    return StdGenericAliasType if isinstance(obj, StdGenericAliasType) else type(obj)


def get_represented_types(
    type_annotation: TypeAnnotation,
    *,
    globalns: Optional[Dict[str, Any]] = None,
    localns: Optional[Dict[str, Any]] = None,
) -> tuple[type, ...]:
    """Return a tuple with all the actual types contained in a type annotation."""

    def recurse_all(annotations: Iterable[TypeAnnotation]) -> tuple[type, ...]:
        return _functools.reduce(lambda acc, c: acc + get_represented_types(c), annotations, ())

    if type_annotation is Ellipsis:
        return ()

    if is_actual_type(type_annotation):
        return (type_annotation,)

    if isinstance(type_annotation, TypeVar):
        if type_annotation.__bound__:
            return get_represented_types(type_annotation.__bound__)
        if type_annotation.__constraints__:
            return recurse_all(type_annotation.__constraints__)
        if typevar_default := getattr(type_annotation, "__default__", None):
            return get_represented_types(typevar_default)

    if isinstance(type_annotation, ForwardRef):
        return get_represented_types(
            eval_forward_ref(type_annotation, globalns=globalns, localns=localns)
        )

    # Generic types
    origin_type = get_origin(type_annotation)
    type_args = get_args(type_annotation)

    if origin_type in [Literal, Union, _types.UnionType]:
        return recurse_all(t for t in type_args)

    if origin_type is not None:
        return (origin_type,)

    return ()


def is_type_with_custom_hash(type_: Type) -> bool:
    return type_.__hash__ not in (None, object.__hash__)


class HasCustomHash(Hashable):
    """ABC for types defining a custom hash function."""

    @classmethod
    def __subclasshook__(cls, candidate_cls: type) -> bool:
        return is_type_with_custom_hash(candidate_cls)


class TypedNamedTupleABC(_abc.ABC, Generic[_T_co]):
    """ABC for `tuple` subclasses created with `collections.abc.namedtuple()`."""

    # Replicate the standard tuple API
    @overload
    @_abc.abstractmethod
    def __getitem__(self, index: int) -> _T_co: ...

    @overload
    @_abc.abstractmethod
    def __getitem__(self, index: slice) -> Self: ...

    @_abc.abstractmethod
    def __getitem__(self, index: Union[int, slice]) -> Union[_T_co, Self]: ...

    @_abc.abstractmethod
    def __len__(self) -> int: ...

    @_abc.abstractmethod
    def __contains__(self, value: object) -> bool: ...

    @_abc.abstractmethod
    def __iter__(self) -> Iterator[_T_co]: ...

    @_abc.abstractmethod
    def __add__(self, other: Self) -> Self: ...

    @_abc.abstractmethod
    def __mul__(self, other: int) -> Self: ...

    @_abc.abstractmethod
    def __rmul__(self, other: int) -> Self: ...

    @_abc.abstractmethod
    def index(self, value: Any, start: int = 0, stop: Optional[int] = None) -> int: ...

    @_abc.abstractmethod
    def count(self, value: Any) -> int: ...

    # Add specific namedtuple methods
    _fields: ClassVar[tuple[str, ...]]

    @_abc.abstractmethod
    def _make(self, iterable: Iterable) -> Self: ...

    @_abc.abstractmethod
    def _asdict(self) -> dict[str, Any]: ...

    @_abc.abstractmethod
    def _replace(self, **kwargs: Any) -> Self: ...

    @classmethod
    def __subclasshook__(cls, subclass: type) -> bool:
        return (
            issubclass(subclass, tuple)
            and (_typing.NamedTuple in getattr(subclass, "__orig_bases__", ()))
        ) or (
            (field_names := getattr(subclass, "_fields", None)) is not None
            and {*field_names} <= _typing.get_type_hints(subclass).keys()
        )


class DataclassABC(_abc.ABC):
    """ABC for data classes."""

    __dataclass_fields__: ClassVar[dict[str, _dataclasses.Field]]
    __dataclass_params__: ClassVar[_DataclassParamsABC]

    @classmethod
    def __subclasshook__(cls, subclass: type) -> bool:
        return _dataclasses.is_dataclass(subclass)


class _DataclassParamsABC(_abc.ABC):
    init: bool
    repr: bool
    eq: bool
    order: bool
    unsafe_hash: bool
    frozen: bool
    match_args: bool
    kw_only: bool
    slots: bool
    weakref_slot: bool


class FrozenDataclass(DataclassABC):
    """ABC for frozen data classes."""

    __dataclass_params__: ClassVar[_FrozenDataclassParamsABC]

    @_abc.abstractmethod
    def __setattr__(self, name: str, value: Any) -> Never: ...

    @classmethod
    def __subclasshook__(cls, subclass: type) -> bool:
        try:
            return _dataclasses.is_dataclass(subclass) and (
                subclass.__dataclass_params__.frozen is True  # type: ignore[attr-defined]  # subclass.__dataclass_params__ is ok after check
            )
        except AttributeError:
            return False


class _FrozenDataclassParamsABC(_DataclassParamsABC):
    frozen: Literal[True]


_KT = TypeVar("_KT", contravariant=True)
_VT = TypeVar("_VT")


class OpaqueMutableMapping(Protocol[_KT, _VT]):
    """
    Mutable mapping without access to the keys, just setting, getting, deleting with a given key.
    """

    def __getitem__(self, key: _KT) -> _VT: ...

    def __setitem__(self, key: _KT, value: _VT) -> None: ...

    def __delitem__(self, key: _KT) -> None: ...


is_protocol = _typing_extensions.is_protocol


def get_partial_type_hints(
    obj: Union[
        object,
        Callable,
        _types.FunctionType,
        _types.BuiltinFunctionType,
        _types.MethodType,
        _types.ModuleType,
        _types.WrapperDescriptorType,
        _types.MethodWrapperType,
        _types.MethodDescriptorType,
    ],
    globalns: Optional[Dict[str, Any]] = None,
    localns: Optional[Dict[str, Any]] = None,
    include_extras: bool = False,
) -> Dict[str, Union[Type, ForwardRef]]:
    """Return a dictionary with type hints (using forward refs for undefined names) for a function, method, module or class object.

    For each member type hint in the object a :class:`typing.ForwardRef` instance will be
    returned if some names in the string annotation have not been found. For additional
    information see :func:`typing.get_type_hints`.
    """
    if getattr(obj, "__no_type_check__", None):
        return {}
    if not hasattr(obj, "__annotations__"):
        return get_type_hints(
            obj, globalns=globalns, localns=localns, include_extras=include_extras
        )

    hints: Dict[str, Union[Type, ForwardRef]] = {}
    annotations = getattr(obj, "__annotations__", {})
    for name, hint in annotations.items():
        obj.__annotations__ = {name: hint}
        try:
            resolved_hints = get_type_hints(
                obj, globalns=globalns, localns=localns, include_extras=include_extras
            )
            hints[name] = resolved_hints[name]
        except NameError as error:
            if isinstance(hint, str):
                # This conversion could be probably skipped in Python versions containing
                # the fix applied in bpo-41370. Check:
                # https://github.com/python/cpython/commit/b465b606049f6f7dd0711cb031fdaa251818741a#diff-ddb987fca5f5df0c9a2f5521ed687919d70bb3d64eaeb8021f98833a2a716887R344
                hints[name] = ForwardRef(hint)
            elif isinstance(hint, (ForwardRef, _typing.ForwardRef)):
                hints[name] = hint
            else:
                raise error

    obj.__annotations__ = annotations

    return hints


def eval_forward_ref(
    ref: Union[str, ForwardRef],
    globalns: Optional[Dict[str, Any]] = None,
    localns: Optional[Dict[str, Any]] = None,
    *,
    include_extras: bool = False,
) -> SolvedTypeAnnotation:
    """Resolve forward references in type annotations.

    Arguments:
        globalns: globals ``dict`` used in the evaluation of the annotations.
        localns: locals ``dict`` used in the evaluation of the annotations.

    Keyword Arguments:
        include_extras: if ``True``, ``Annotated`` hints will preserve the annotation.

    Examples:
        >>> from typing import Dict, Tuple
        >>> print("Result:", eval_forward_ref("Dict[str, Tuple[int, float]]"))
        Result: ...ict[str, ...uple[int, float]]

    """

    def f() -> None: ...

    f.__annotations__ = {"return": ForwardRef(ref) if isinstance(ref, str) else ref}

    safe_localns = {**localns} if localns else {}
    safe_localns.setdefault("typing", _sys.modules[__name__])
    safe_localns.setdefault("NoneType", type(None))

    actual_type = get_type_hints(f, globalns, safe_localns, include_extras=include_extras)["return"]
    assert not isinstance(actual_type, ForwardRef)

    return actual_type


def _collapse_type_args(*args: Any) -> Tuple[bool, Tuple]:
    if args and all(args[0] == a for a in args[1:]):
        return (True, args)
    else:
        return (False, args)


@final
@_dataclasses.dataclass
class CallableKwargsInfo:
    data: Dict[str, Any]


def infer_type(
    value: Any, *, annotate_callable_kwargs: bool = False, none_as_type: bool = True
) -> TypeAnnotation:
    """Generate a typing definition from a value.

    Keyword Arguments:
        annotate_callable_kwargs: if ``True``, ``Callable``s will be returned as
            a ``Annotated[Callable, CallableKwargsInfo]`` hint, where :class:`CallableKwargsInfo`
            contains the inferred typings for the keyword arguments, if any.
        none_as_type:  if ``True``, ``None`` hints will be transformed to ``type(None)``.

    Examples:
        >>> infer_type(3)
        <class 'int'>

        >>> infer_type((3, "four"))
        tuple[int, str]

        >>> infer_type((3, 4))
        tuple[int, int]

        >>> infer_type(frozenset([1, 2, 3]))
        frozenset[int]

        >>> infer_type({"a": 0, "b": 1})
        dict[str, int]

        >>> infer_type({"a": 0, "b": "B"})
        dict[str, ...Any]

        >>> print("Result:", infer_type(lambda a, b: a + b))
        Result: ...Callable[[...Any, ...Any], ...Any]

        # Note that some patch versions of cpython3.9 show weird behaviors
        >>> def f(a: int, b) -> int: ...
        >>> print("Result:", infer_type(f))
        Result: ...Callable[[...int..., ...Any], int]

        >>> def f(a: int, b) -> int: ...
        >>> print("Result:", infer_type(f))
        Result: ...Callable[..., int]

        >>> print("Result:", infer_type(Dict[int, Union[int, float]]))
        Result: ...ict[int, ...int...float...]

    For advanced cases, using :func:`functools.singledispatch` with custom hooks
    is a simple way to extend and customize this base implementation.

    Example:
        >>> import functools, numbers
        >>> extended_infer_type = functools.singledispatch(infer_type)
        >>> @extended_infer_type.register(int)
        ... @extended_infer_type.register(float)
        ... @extended_infer_type.register(complex)
        ... def _infer_type_number(value, *, annotate_callable_kwargs: bool = False):
        ...     return numbers.Number
        >>> extended_infer_type(3.4)
        <class 'numbers.Number'>
        >>> infer_type(3.4)
        <class 'float'>

    """
    _infer = _functools.partial(infer_type, annotate_callable_kwargs=annotate_callable_kwargs)

    if isinstance(value, (StdGenericAliasType, _TypingSpecialFormType)):
        return value

    if value in (None, type(None)):
        return type(None) if none_as_type else None

    if isinstance(value, type):
        return Type[value]

    if isinstance(value, tuple) and not isinstance(value, TypedNamedTupleABC):
        # Special case for tuples, which can have multiple types.
        # We should not confuse them with namedtuples, which are
        # treated as normal classes.
        _, args = _collapse_type_args(*(_infer(item) for item in value))
        if args:
            return StdGenericAliasType(tuple, args)
        else:
            return StdGenericAliasType(tuple, (Any, ...))

    if isinstance(value, (list, set, frozenset)):
        t: Union[Type[List], Type[Set], Type[FrozenSet]] = type(value)
        unique_type, args = _collapse_type_args(*(_infer(item) for item in value))
        return StdGenericAliasType(t, args[0] if unique_type else Any)

    if isinstance(value, dict):
        unique_key_type, keys = _collapse_type_args(*(_infer(key) for key in value.keys()))
        unique_value_type, values = _collapse_type_args(*(_infer(v) for v in value.values()))
        kt = keys[0] if unique_key_type else Any
        vt = values[0] if unique_value_type else Any
        return StdGenericAliasType(dict, (kt, vt))

    if isinstance(value, _types.FunctionType):
        try:
            annotations = get_type_hints(value)
            return_type = annotations.get("return", Any)

            sig = _inspect.signature(value)
            arg_types: List = []
            kwonly_arg_types: Dict[str, Any] = {}
            for p in sig.parameters.values():
                if p.kind in (
                    _inspect.Parameter.POSITIONAL_ONLY,
                    _inspect.Parameter.POSITIONAL_OR_KEYWORD,
                ):
                    arg_types.append(annotations.get(p.name, None) or Any)
                elif p.kind == _inspect.Parameter.KEYWORD_ONLY:
                    kwonly_arg_types[p.name] = annotations.get(p.name, None) or Any
                elif p.kind in (
                    _inspect.Parameter.VAR_POSITIONAL,
                    _inspect.Parameter.VAR_KEYWORD,
                ):
                    raise TypeError("Variadic callables are not supported")

            result: Any = Callable[arg_types, return_type]
            if annotate_callable_kwargs:
                result = Annotated[result, CallableKwargsInfo(kwonly_arg_types)]
            return result
        except Exception:
            return Callable

    return type(value)


# TODO(egparedes): traversing a typing definition is an operation needed in several places
#   but it currently requires custom and cumbersome code due to the messy implementation details
#   in the standard library. Ideally, this code could be replaced by translating it once to a
#   custom "typing tree" data structure which could be then traversed in a generic way.
#



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import List

from gt4py import eve
from gt4py.cartesian.gtc import oir
from gt4py.cartesian.gtc.passes.oir_optimizations.utils import AccessCollector


class MaskStmtMerging(eve.NodeTranslator):
    def _merge(self, stmts: List[oir.Stmt]) -> List[oir.Stmt]:
        merged = [self.visit(stmts[0])]
        for stmt in stmts[1:]:
            stmt = self.visit(stmt)
            if (
                isinstance(stmt, oir.MaskStmt)
                and isinstance(merged[-1], oir.MaskStmt)
                and stmt.mask == merged[-1].mask
                and not (
                    AccessCollector.apply(merged[-1].body).write_fields()
                    & AccessCollector.apply(stmt.mask, is_write=False).fields()
                )
            ):
                merged[-1] = oir.MaskStmt(mask=merged[-1].mask, body=merged[-1].body + stmt.body)
            else:
                merged.append(stmt)
        return merged

    def visit_HorizontalExecution(self, node: oir.HorizontalExecution) -> oir.HorizontalExecution:
        return oir.HorizontalExecution(
            body=self._merge(node.body), declarations=node.declarations, loc=node.loc
        )

    # Stmt node types with lists of Stmts within them:

    def visit_MaskStmt(self, node: oir.MaskStmt) -> oir.MaskStmt:
        return oir.MaskStmt(mask=node.mask, body=self._merge(node.body), loc=node.loc)

    def visit_While(self, node: oir.While) -> oir.While:
        body_nodes = []
        for stmt in node.body:
            if isinstance(stmt, oir.MaskStmt) and node.cond == stmt.mask:
                body_nodes.extend(stmt.body)
            else:
                body_nodes.append(stmt)
        return oir.While(cond=self.visit(node.cond), body=self.visit(body_nodes), loc=node.loc)
